from __future__ import print_function
#
#   collection of widgets for artists
#
import wx
import weakref
import numpy as np
from ifigure.utils.edit_list import EditListPanel, EDITLIST_CHANGED, ScrolledEditListPanel
from ifigure.utils.cbook import FindFrame

from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod

from ifigure.mto.fig_text import FigText
from ifigure.mto.fig_axes import FigAxes
from ifigure.mto.fig_axes import FigInsetAxes
from ifigure.mto.fig_axspan import FigAxspan
from ifigure.mto.fig_axline import FigAxline
from ifigure.mto.fig_axlinec import FigAxlineC
from ifigure.mto.fig_plot import FigPlot
from ifigure.mto.fig_contour import FigContour
from ifigure.mto.fig_image import FigImage
from ifigure.mto.fig_page import FigPage
from ifigure.mto.fig_legend import FigLegend
from ifigure.mto.fig_axes import FigColorBar

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('ArtistWidgets')

# definition of parameter for edit_list
# listparam[0:3] : parameters to be sent to editlist
# listparam[0]   : label
# listparam[1]   : value
# listparam[2]   : mode
# listparam[3]   : setting
# listparam[4]   : property name
# listparam[5]   :
#    flag : -1: do nothing
#            0: use set_xxx of artist
#           10: use set_xxx of artist (applied to all artists)
#            1: use setp/getp of FigObj
#            2: call method (set_xxx/get_xxx) of FigObj
#               xxx is property name. these methods should
#               have the following signature
#               set_xxx(self, value, artist)
#               get_xxx(self, artist)
#            3: call method (set_xxx/get_xxx) of FigObj
#               extension of flag = 2
#               set_xxx(self, (tab, value), artist)
#               get_xxx(self, artist, tab)
#           12: the same as 2, but bypass history
#           13: the same as 3, but bypass history
# listparam[6] = menu_name
listparam = {}
#s = {"minV": 0.,"maxV": 359., "val" : 90, "res" : 1, "text_box" : False}
s_fontsize = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
              "choices": ["7", "8", "9", "10", "11", "12", "14",
                          "16", "18", "20", "22", "24", "26", "28",
                          "36", "48", "72"]}
s_fontsize2 = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
               "choices": ["default", "7", "8", "9", "10", "11", "12", "14",
                           "16", "18", "20", "22", "24", "26", "28",
                           "36", "48", "72"]}
s_markersize = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
                "choices": ["5", "6", "7", "8", "9", "10", "11", "12", "14",
                            "16"]}
s_markeredgewidth = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
                     "choices": ["0.5", "1", "2", "3", "4", "5"]}
s_imageinterp = {"style": wx.CB_READONLY,
                 "choices": ["nearest", "linear", "cubic"]}
s_shading = {"style": wx.CB_READONLY,
             "choices": ["flat", "gouraud"]}
s_axxpos = {"style": wx.CB_READONLY, "choices": ["bottom", "top"]}
s_axypos = {"style": wx.CB_READONLY, "choices": ["left", "right"]}
s_splineinterp = {"style": wx.CB_DROPDOWN,
                  "choices": ["linear", "spline"]}
s_multialignment = {"style": wx.CB_DROPDOWN,
                    "choices": ["left", "center", "right"]}


###
listparam["multialignment"] = ["alignment", "left",
                               4,  s_multialignment, 'multialignment', 2]
listparam["spinterp"] = ["interpolation", "spline",
                         4, s_splineinterp, 'sp_interp', 2]
listparam["axxpos"] = ["position", ("bottom", True, True, None),
                       26, s_axxpos, 'ax_pos', 3, 'axpos']
listparam["axypos"] = ["position", ("left", True, True, None),
                       26, s_axypos, 'ax_pos', 3, 'axpos']

listparam["rotation"] = ["rotation", '0',   25,  {}, 'rotation', 0]
listparam["text"] = ["text",         '',  200, {}, 'text', 0]
listparam["legendtitle"] = ["text",         '',  200, {}, 'title', 2]
s = {"text": 'shadow'}
listparam["legendshadow"] = [None,   False,  3, s, 'shadow', 2]

listparam["color"] = ["color",      'red',  6, {}, 'color', 0]
#listparam["linecolor"] = ["color",  'red', 406, {}, 'color', 0]
listparam["ecolor"] = ["color",      'red',  6, {}, 'ecolor', 2]
listparam["alpha"] = ["alpha",       1.0, 105, {}, 'alpha', 0]


listparam["facecolor"] = ["fill",      'red', 106, {}, 'facecolor', 0]
listparam["edgecolor"] = ["edge",      'red',  6, {}, 'edgecolor', 0]
listparam["pedgecolor_2"] = ["edge",     'red', 306, {},  'edgecolor', 2]

listparam["xtcolor"] = ["tick",      'red',  6, {}, 'xtcolor', 2]
listparam["ytcolor"] = ["tick",      'red',  6, {}, 'ytcolor', 2]
listparam["xlcolor"] = ["label",      'red',  6, {}, 'xlcolor', 2]
listparam["ylcolor"] = ["label",      'red',  6, {}, 'ylcolor', 2]

listparam["linewidth"] = ["line width", 2.,  7, {}, 'linewidth', 0]
listparam["linewidthz_0"] = ["line width", 2.,  107, {}, 'linewidth', 0]
listparam["linewidthz"] = ["line width", 2.,  107, {}, 'linewidth', 2]
listparam["elinewidth"] = ["line width", 2.,  7, {}, 'elinewidth', 2]
listparam["linestyle"] = ["line style", 'red',  8, {}, 'linestyle', 0]
listparam["marker"] = ["marker",      'red',  9, {}, 'marker',  0]
listparam["markerfacecolor"] = ["fill",  'red', 106, {}, 'markerfacecolor', 0]
listparam["markeredgecolor"] = ["edge",  'red',  6, {}, 'markeredgecolor', 0]
listparam["markersize"] = ["size",  "6",  104, s_markersize, 'markersize', 0]
listparam["markeredgewidth"] = [
    "edge width",  2.,  7, {}, 'markeredgewidth', 0]
listparam["plinestyle"] = ["line style", 'dotted',  10, {}, 'linestyle', 0]

s = {"text": 'fill'}
listparam["fill"] = ["fill area", True,  3, s, 'fill', 0]
s = {"text": 'closed'}
listparam["closepoly"] = ["closed", True,  3, s, 'closepoly', 2]
s = {"minV": 0., "maxV": 30., "val": 10, "res": 1, "text_box": False}
listparam["contour_nlevel"] = ["level", '0',   24,  s, 'contour_nlevel', 2]
s["expand_space"] = 1
mmm = [[None, '0',   24,  s], ]
mm2 = [[None, '0',  100,  s], ]
listparam["contour_nlevel2"] = ["levels",  (False, ['7'], ['10']), 27,
                                ({"text": 'use expression'},
                                 {"elp": mm2}, {"elp": mmm}),
                                "contour_nlevel2", 2]
mm21 = [['   ',   'red', 106, {}], ]
mm2 = [['size', '10',  104,  s_fontsize],
       [None, (False, ['red']), 27,
        ({"text": 'use fixed color'}, {"elp": mm21})],
       ['inline', True,  3, {'text': 'on'}],
       ['spaceing', '5', 24,
        {"minV": 1., "maxV": 30., "val": 5, "res": 1, "text_box": False}],
       ['format', '%1.3f',   0, {}],
       ['skip', '0', 24,
        {"minV": 0., "maxV": 5., "val": 0, "res": 1, "text_box": True}],
       ["alpha",  1.0 , 105, {}, ],
       ]
init_v = [8.0, (False, [(0.0, 0.0, 0.0, 1.0)]), True, 5.0, u'%1.3f', '0',  1.0]
s_decimate = {"text": 'on'}
listparam['decimate'] = ["decimate", True,  3, s_decimate, 'decimate', 2]

listparam['clabel_param'] = [None,  (False, init_v), 27,
                             ({"text": 'use labels'}, {"elp": mm2}),
                             "clabel_param", 2]
#listparam["contour_alpha"] =["alpha",  1.0 , 105, {}, 'alpha', 1]
s = {"text": 'fill'}
listparam["contour_fill"] = ["fill area", True,  3, s, 'contour_fillmode', 2]
s = {"text": 'shade'}
listparam["surf_shade"] = ["shade", True,  3, s, 'shade', 2]
s = {"minV": 5., "maxV": 30., "val": 12, "res": 1, "text_box": False}
listparam["size"] = ["size", "12",  104,  s_fontsize, 'size', 0]
listparam["textsize"] = ["size", "12",  104,  s_fontsize, 'size', 1]
listparam["xlsize"] = ["size", "12",  104,  s_fontsize, 'xlsize', 2]
listparam["ylsize"] = ["size", "12",  104,  s_fontsize, 'ylsize', 2]

listparam["cmap"] = ["color map", 3, 12, {}, 'cmap', 2]
listparam["cmap3"] = ["color map", 3, 12, {}, 'cmap3', 3, 'cmap']

listparam["caxis"] = ["color axis", 'c', 28, None, 'selected_caxis', 2]
listparam["image_interp"] = ["interpolation", "linear",
                             4, s_imageinterp, 'interp', 2]
listparam["tripcolor_shading"] = ["interpolation", "flat",
                                  4, s_shading, 'shading', 1]

s_shading2 = {"style": wx.CB_READONLY,
              "choices": ["flat", "linear"]}
listparam["solid_shade"] = ["interpolation", "flat", 4, s_shading2,
                            "shade", 2]
s = {"values": ['data', 'axes']}
listparam["switchtrans"] = ["coord",        0, 1, s, 'switchtrans', 2]
s = {"values": ['data', 'axes', 'figure'], "orientation": 'vertical'}
listparam["inset_switchtrans"] = ["coord",        0, 1, s, 'switchtrans', 2]
listparam["inset_size"] = ["size(x, y, w, h)",     '', 0, {}, 'insetsize', 2]
listparam["title"] = ["title",        None,  115, {}, 'title', 2]
s = {"values": ['auto', 'equal']}
listparam["aspect"] = ["aspect",        0,  1, s, 'aspect', 2]
s = {"text": 'suppress clipping by 3D cut-plane'}
listparam["noclip3d"] = [None,        0,  3, s, 'noclip3d', 1]
s = {"text": 'show'}
listparam["frame"] = ["frame",        0,  3, s, 'frame_on', 0]
listparam["axis"] = ["axis",        0,  3, s, 'axis_onoff', 2]
listparam["xlabel"] = ["label",      None,   15, {}, 'xlabel', 2]
listparam["ylabel"] = ["label",      None,   15, {}, 'ylabel', 2]

listparam["xscale"] = ["scale", 'linear',  14,  None, 'xscale', 2]
listparam["yscale"] = ["scale", 'linear',  14,  None, 'yscale', 2]
listparam["xrange"] = ["range", ('0', '1'),  13,  None, 'xlim', 2]
listparam["yrange"] = ["range", ('0', '1'),  13,  None, 'ylim', 2]
s = {"text": 'on'}
listparam["fancybox"] = ["fancybox", True,  3, s, 'fancybox', 1]
listparam["fbalpha"] = ["alpha",      1.0, 105, {}, 'fbalpha', 1]
listparam["fbcolor"] = ["face",      'red',  6, {}, 'fbcolor', 1]
listparam["fbecolor"] = ["edge",   'red',  6, {}, 'fbecolor', 1]
s = {"style": wx.CB_READONLY,
     "choices": ["larrow", "rarrow", "round", "round4",
                 "roundtooth", "sawtooth", "square"]}
listparam["fbstyle"] = ["shape", 'square',  4, s, 'fbstyle', 1]
s = {"minV": 0.1, "maxV": 0.7, "val": 0.3, "res": 0.05, "text_box": False}
listparam["fbpad"] = ["pad", 0.3,   24,  s, 'fbpad', 1]
listparam["arrowstyle"] = ["", 'simple',   16,  None, 'arrowstyle', 2]
listparam["mdssource"] = [None, None,   18,  None, 'mdssource', 2]
s = {"style": wx.CB_READONLY,
     "choices": ["plot", "contour", "image", "axline", "axspan"]}
listparam["mdsfiguretype"] = [None,  'plot',  31, s, 'mdsfiguretype', 12]
listparam["mdsrange"] = [None,  ((False, (-1, 1)),
                                 (False, (-1, 1))),
                         32, None, 'mdsrange', 2]
#listparam["mdstxt"]   = [None,   'This is a MDSplus data object',  2,   {}, None, -1]
listparam["mdsevent"] = ["update",   '',  0,   {}, 'mdsevent',  2]
s = {"style": wx.CB_READONLY,
     "choices": ["circle", "ellipse"]}
listparam["circle_type"] = ["type", 'circle',  4, s, 'circletype', 2]
s = {"style": wx.TE_PROCESS_ENTER,
     "choices": ["0", "30", "45", "60", "90", "120", "135",
                 "150", "180"]}
listparam["circle_angle"] = ["angle", 0,  104, s, 'circleangle', 2]
s = {"style": wx.CB_READONLY,
     "choices": ["square", "rectangle"]}
listparam["box_type"] = ["type", 'square',  4, s, 'boxtype', 2]

#listparam["legendlabelcolor"] = ["color",  'red',  6, {}, 'legendlabelcolor', 3]
#listparam["legendlabel"]  = ["text",         '',   0, {}, 'legendlabel', 3]
s1 = {"style": wx.CB_READONLY,
      "choices": ["serif", "sans-serif",
                  "cursive", "fantasy", "monospace", "default"]}
s2 = {"style": wx.CB_READONLY,
      "choices": ["ultralight", "light", "normal",
                  "regular", "book", "medium",
                  "roman", "semibold", "demibold",
                  "demi", "bold", "heavy",
                  "extra bold", "black", "default"]}
s3 = {"style": wx.CB_READONLY,
      "choices": ["normal", "italic", "oblique", "default"]}
s4 = {"style": wx.TE_PROCESS_ENTER,
      "choices": ["5", "7", "8", "9", "12", "15",
                  "18", "20", "24", "36", "48"]}


listparam["cxclipunder"] = ["Under", (False, 'white'), 38,  ({"text": "Clip"}, {}),
                            'cxclipunder', 3, 'clipunder']
listparam["cxclipover"] = ["Over",  (False, 'white'),   38, ({"text": "Clip"}, {}),
                           'cxclipover',  3, 'clipover']

s = {"elp": (("text",   '',     0, {}),
             ("color",  'red',  6, {}),)}

ss = {"elp": (("font",   'default',  4, s1,),
              ("weight", 'default',  4, s2,),
              ("style",  'default',  4, s3,),
              ("size",     10,      104,  s4),)}

listparam["legendloc"] = [None, '',   19,  None, 'legendloc', 2]
listparam["legendentry"] = [None,  (-1, ('1',),
                                    (('', 'red'),)),
                            227, s, 'legendentry', 2]
listparam["legendlabelprop"] = [None,
                                ('default', 'default', 'default', 12),
                                33, ss, 'legendlabel_prop', 2]


s_spec_nfft = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
               "choices": ["64", "128", "256", "512", "1024", "2048", "4096",
                           "8192", "16384"]}
s_spec_nfft16 = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
                 "choices": ["16", "32", "64", "128", "256", "512", "1024", "2048", "4096",
                             "8192", "16384"]}
s_spec_nfft_none = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
                    "choices": ["none", "64", "128", "256", "512", "1024", "2048", "4096",
                                "8192", "16384"]}
s_spec_window = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
                 "choices": ["none", "hanning", "hamming", "blackman", "bartlett"]}
s_spec_sided = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
                "choices": ["onesided", "twosided"]}
s_spec_detrend = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
                  "choices": ["none", "mean", "linear"]}
elp_spec_fftp = [["NFFT", "256",  104,  s_spec_nfft],
                 ["PadTo", "256",  104,  s_spec_nfft_none]]
listparam["spec_fftp"] = [None, ("256", "256"),  33,  {
    "elp": elp_spec_fftp}, 'spec_fftp', 2]
listparam["spec_noverlap"] = ["Overlap", "128",
                              104,  s_spec_nfft16, 'spec_noverlap', 2]
listparam["spec_window"] = ["Window", "hanning",
                            4,  s_spec_window, 'spec_window', 2]
listparam["spec_sided"] = ["Sides", "onesided",
                           4,  s_spec_sided, 'spec_sides', 2]
listparam["spec_detrend"] = ["Detrend", "none",
                             4,  s_spec_detrend, 'spec_detrend', 2]

listparam["axrangeparam13"] = [None,    None,
                               20, {}, 'axrangeparam13', 13, 'axrangeparam']
listparam["axrangeparam13o"] = [None,    None,   20, {'check_range_order': True},
                                'axrangeparam13', 13, 'axrangeparam']
listparam["axcrangeparam13"] = [None,    None,   20, {'check_range_order': True},
                                'axcrangeparam13', 13, 'axcrangeparam']

s2 = {"style": wx.CB_DROPDOWN | wx.TE_PROCESS_ENTER,
      "choices": ["default", "scalar", "scalar(mathtext)", "log",  "log(mathtext)", "log(exp)",
                  "none"]}

listparam["axformat"] = ["format", "default",  4,  s2, 'axformat', 3]
listparam["axlabel"] = ["label",      None,  115, {}, 'axlabel', 3]
listparam["axlsize"] = ["size", "12",   104,  s_fontsize2, 'axlsize', 3]
listparam["axotsize"] = ["size", "12",   104,  s_fontsize2, 'axotsize', 3]
listparam["axlotsize"] = ["size", "12",   42,  s_fontsize2, 'axlotsize', 3]
listparam["axtcolor"] = ["tick", 'red', 206, {}, 'axtcolor', 3]

listparam["axlcolor"] = ["label", None, 506, {}, 'axlcolor', 3]
listparam["axtlcolor"] = [None,  None, 606, {}, 'axtlcolor', 3]
listparam["axticks"] = ["tick loc", 'Auto', 37, {}, 'axticks', 3]

listparam["axis_bgedgenone"] = [None, "background",  2,  {}, 'none',  -1]
listparam["axis_bgcolor"] = ["face", 'red', 206, {}, 'axis_bgfacecolor', 2]
listparam["axis_bgedgecolor"] = [
    "edge",   'red', 206, {}, 'axis_bgedgecolor', 2]
listparam["axis_bglinewidth"] = [
    "width",     0, 107, {}, 'axis_bglinewidth', 2]
listparam["axis_bglinestyle"] = [
    "style",   'solid', 10, {}, 'axis_bglinestyle', 2]
listparam["axis_bgalpha"] = ["alpha",       1.0, 105, {}, 'axis_bgalpha', 2]
listparam["axis3d_bgcolor"] = ["b.g.(3D)", ('grey', 'grey', 'grey'),
                               23, {}, 'axis3d_bgcolor', 2]
listparam["axis3d_bgalpha"] = ["alpha(3D)", "1",
                               105, {}, 'axis3d_bgalpha', 2]
s1 = {"style": wx.CB_READONLY,
      "choices": ["serif", "sans-serif",
                  "cursive", "fantasy", "monospace"]}
s2 = {"style": wx.CB_READONLY,
      "choices": ["ultralight", "light", "normal",
                  "regular", "book", "medium",
                  "roman", "semibold", "demibold",
                  "demi", "bold", "heavy",
                  "extra bold", "black"]}
s3 = {"style": wx.CB_READONLY,
      "choices": ["normal", "italic", "oblique"]}
listparam["fontfamily"] = ["font",    s1["choices"][0], 4,
                           s1, 'fontfamily', 1]
listparam["fontweight"] = ["weight",  s2["choices"][0], 4,
                           s2, 'fontweight', 1]
listparam["fontstyle"] = ["style",   s3["choices"][0], 4,
                          s3, 'fontstyle',  1]
listparam["epsscale"] = ["scale",   (True, True, 100, 100), 29,
                         None, 'epsscale',  2]

listparam["anchor"] = ["anchor",   (0, 0), 30,  None, 'anchor',  2]
listparam["arrowstyle"] = ["", 'simple',   16,  None, 'arrowstyle', 2]

sscale_setting = {"minV": 0.1, "maxV": 20., "val": 1.0,
                  "res": 0.01, 'motion_event': True}
listparam["scatter_sscale"] = ["size", 1.0, 124, sscale_setting, 'sscale', 2]

mm = [[None, 'simple',  16, None], ]
listparam["curvearrow1"] = [None,  (False, ['simple']), 27,
                            ({"text": 'show arrow1'}, {"elp": mm}), 'curvearrow1', 2]
listparam["curvearrow2"] = [None,  (False, ['simple']), 27,
                            ({"text": 'show arrow2'}, {"elp": mm}), 'curvearrow2', 2]
# make "_2" version
name_ones = ("color", "linestyle", "linewidth",
             "marker", "alpha", "fill", "facecolor",
             "edgecolor", "plinestyle",
             "markerfacecolor", "markeredgecolor",
             "markersize", "markeredgewidth")
# for quivers
s3 = {"style": wx.CB_READONLY,
      "choices": ["xy", "uv"]}
listparam["qangles"] = ["angles",    s1["choices"][0], 4, s3, 'angles',  2]
s = {"minV": 1., "maxV": 8., "val": 5, "res": 1, "text_box": False}
listparam["qheadlength"] = ["headlength", '1',   300,  {}, 'headlength', 2]
listparam["qheadaxislength"] = [
    "headaxislength", '1',   24,  s, 'headaxislength', 2]
listparam["qheadwidth"] = ["headwidth", '1',   24,  s, 'headwidth', 2]

spivot = {"style": wx.CB_READONLY,
          "choices": ["tail", "mid", "tip"]}
listparam["qpivot"] = ["pivot", spivot['choices'][0],   4,
                       spivot, 'pivot', 2]
listparam["q3dlength"] = ["length", 1.,   300, {}, 'q3dlength', 2]
listparam["q3dratio"] = ["length ratio", '0.3',   24,
                         {"minV": 0.05, "maxV": 0.9, "val": 0.3, "res": 0.01,
                          "text_box": False}, 'q3dratio', 2]

# page and axes properties.
# property editor treat these differently.
listparam["bgcolor"] = ["color", 'white', 6, {}, 'facecolor', 0]
# 8 is used in property command and call with nodelete=True
listparam["page_title_size"] = ["titlesize", 12, 104, {}, 'title_size', 8]
listparam["page_ticklabel_size"] = [
    "ticklabel_size", 12, 104, {}, 'ticklabel_size', 8]
listparam["page_axeslabel_size"] = [
    "axeslabel_size", 12, 104, {}, 'axeslabel_size', 8]
listparam["page_axesbox_width"] = [
    "axesbox_width", 1, 104, {}, 'axesbox_width', 8]
listparam["page_axestick_width"] = [
    "axestick_width", 1, 104, {}, 'axestick_width', 8]

listparam["gl_lighting"] = [None,   None, 40,  None, 'gl_lighting',  2]
listparam["gl_view"] = [None,   None, 44,  None, 'axes3d_viewparam',  2]

for name in name_ones:
    v = [x for x in listparam[name]]
    v[5] = 2
    listparam[name + "_2"] = v

# add default undo/redo menu name
# note : this menu name is used also in property commannd
for key in listparam:
    if len(listparam[key]) == 6:
        if key[-2:] == '_1':
            name = key[:-2]
        elif key[-2:] == '_2':
            name = key[:-2]
        else:
            name = key
        listparam[key].append(name)


def call_getter(artist, ll, tab=''):
    sw = ll[5]
#             sw = self.list[k][i][5]
    if sw > 9:
        sw = sw-10
    if sw == 0:
        try:
            m = getattr(artist, 'get_'+ll[4])
            return m()
        except AttributeError:
            return None
        except:
            import traceback
            traceback.print_exc()
            return None
#             elif sw == 10:
#                m = getattr(artist, 'get_'+self.list[k][i][4])
#                value[i] = m()
    elif sw == 1:
        return artist.figobj.getp(ll[4])
    elif sw == 2:
        m = getattr(artist.figobj, 'get_'+ll[4])
        return m(artist)
    elif sw == 3:
        m = getattr(artist.figobj, 'get_'+ll[4])
        return m(artist, tab)
######


class base_artist_widget(object):
    def __init__(self, target_type=1):
        self.target_artist = None
        self.target_artist_mul = None   # multiple selection
        self.target_type = target_type  # 1: first aritst 0: any artist
        self.target_figobj = None
        self.target_figobj_type = ''

    def set_value(self,  artist=None, value=None):
        # this shouuld define how to update
        # gui based on the data in ifigure_canvas
        pass

        #   def get_value(self,  artist=None):
        # pass
        # this shouuld define how to update

        # ifigure_canvas based on GUI

    def set_artist_property(self, evt):
        # routine to set property of artist based on
        # evt happend on aritst_widgets
        print(("debug", "set_artist_property", self))
        pass

    def set_target_artist(self, a, td=None):
        self.target_artist = weakref.ref(a)
        if td is None:
            td = a.figobj
        self.target_figobj = weakref.ref(td)

    def set_target_artist_mul(self, sels):
        if len(sels) > 1:
            self.target_artist_mul = sels
        else:
            self.target_artist_mul = None

    def get_target_artist(self):
        if self.target_artist is None:
            return None
        return self.target_artist()

    def replace_target_artist(self, evt):

        a1 = evt.a1
        a2 = evt.a2

#       print  'here', a1, a2, self.target_type, self
        td = evt.GetTreeDict()
        if self.target_type == 1:
            if td is self.target_figobj():
                self.set_target_artist(a2, td)
#            self.target_artist() = a2
#            self.target_figobj = td
        elif self.target_type == 0:
            if self.target_artist() is a1:
                self.set_target_artist(a2, td)
#            self.target_artist = a2
#            self.target_figobj = td

    def onEL_Changed(self, evt):
        evt.artist_panel = self
        evt.Skip()


class artist_panel(wx.Panel):
    #  define setter/getter for a panel made by
    #  two editlist panels
    #   def __init__(self, *args, **kargs):
    #       print 'artist_panel__init__'
    #       wx.Panel.__init__(self, *args, **kargs)
    def make_panel(self):
        self.target_artist = None
        self.elp = [None]*len(self.list)
        if (self.tab[0]) != '':
            nb = wx.Notebook(self)
            self.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, self.onNBChanged)

        else:
            nb = wx.Panel(self, wx.ID_ANY)
            nb.SetSizer(wx.BoxSizer(wx.VERTICAL))

        i = 0

        for list in self.list:
            self.elp[i] = ScrolledEditListPanel(nb, [x[0:4] for x in list])
            # self.elp[i].SetScrollRate(0,5)
            if (self.tab[0]) != '':
                #          if len(self.list) != 1 :
                nb.AddPage(self.elp[i], self.tab[i])
                # self.elp[i].SetScrollRate(0,5)
                # self.elp[i].SetupScrolling()
            else:
                nb.GetSizer().Add(self.elp[i], 1,
                                  wx.EXPAND, 0)
            i = i+1
        if (self.tab[0]) != '':
            nb.ChangeSelection(0)
#       if len(self.list) > 1: nb.ChangeSelection(0)
        pansizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(pansizer)
        pansizer.Add(nb, 1, wx.EXPAND | wx.ALL,
                     0)
        self.Layout()
        self.Bind(EDITLIST_CHANGED, self.onEL_Changed)
        self.Layout()
        self.Enable(False)
        self.nb = nb

    def Enable(self, value):
        wx.Panel.Enable(self, value)
        for elp in self.elp:
            elp.Enable(value)

    def build_editlist_list(self, obj, ret):
        self.target_figobj_type = type(obj)
        if isinstance(ret, tuple):
            tab = ret[0]
            props = ret[1]
        else:
            tab = ['']
            props = [ret]

        ret = []

        for plist in props:
            l = []
            for p in plist:
                l.append(listparam[p])
            ret.append(l)
#       if tab is None:
#          return ret
#       else:
        return tab, ret

    def set_value(self, artist):
        #
        #   artist/figobj -> palette
        #
        if not isinstance(artist.figobj, self.target_figobj_type):
            return

        self.adjust_elp(artist)
        for k in range(len(self.list)):
            if not self.elp[k].IsEnabled():
                continue
            value = ['']*len(self.list[k])
            for i in range(len(value)):
                if self.list[k][i][4] == 'none':
                    value[i] = self.list[k][i][1]
                    continue

                if len(self.tab) > k:
                    value[i] = call_getter(
                        artist, self.list[k][i], self.tab[k])
                else:
                    value[i] = call_getter(artist, self.list[k][i])

                sw = self.list[k][i][5]
                if sw > 9:
                    sw = sw-10
                if sw == 0:
                    m = getattr(artist, 'get_'+self.list[k][i][4])
                    value[i] = m()
                elif sw == 1:
                    value[i] = artist.figobj.getp(self.list[k][i][4])
                elif sw == 2:
                    m = getattr(artist.figobj, 'get_'+self.list[k][i][4])
                    value[i] = m(artist)
                elif sw == 3:
                    m = getattr(artist.figobj, 'get_'+self.list[k][i][4])
                    value[i] = m(artist, self.tab[k])
            self.elp[k].SetValue(value)

    def set_artist_property(self, evt):
        #
        #   palette -> artist/figobj
        #
        #   artist/figobj will be modified in an
        #   undoalbe way using undo_redo_history
        #
        from ifigure.utils.cbook import isDescendant

        if self.target_artist is None:
            return None, None
        if self.target_artist() is None:
            return None, None
        if self.target_artist().figobj is None:
            self.target_artist = None
            return None, None
        c = 0
        for elp in self.elp:
            if isDescendant(elp, evt.GetEventObject()):
                k = c
                break
            c = c+1

        value = self.elp[k].GetValue()
        i = evt.widget_idx

        name = self.list[k][i][4]
        tab = self.tab[k]
        v = value[i]
        actions = []
        targets = ([self.target_artist] if self.target_artist_mul is None
                   else self.target_artist_mul)
        if self.list[k][i][5] == 0:
            for t in targets:
                action = UndoRedoArtistProperty(t(), name, v)
                actions.append(action)

        elif self.list[k][i][5] == 1:
            for t in targets:
                action = UndoRedoFigobjProperty(t(), name, v)
                actions.append(action)
        elif self.list[k][i][5] == 2:
            for t in targets:
                action = UndoRedoFigobjMethod(t(), name, v)
                actions.append(action)
        elif self.list[k][i][5] == 3:
            action = UndoRedoFigobjMethod(self.target_artist(),
                                          name, v)
            action.set_extrainfo(tab)
            actions.append(action)
        elif self.list[k][i][5] == 12:
            # print self.target_artist()
            m = getattr(self.target_artist().figobj, 'set_'+name)
            m(v, self.target_artist())
            return None, None
        elif self.list[k][i][5] == 13:
            m = getattr(self.target_artist().figobj, 'set_'+name)
            m((tab, v), self.target_artist())
            return None, None
        return actions, self.list[k][i][6]

    def add_action_to_history(self, evt, actions, name=''):
        window = evt.GetEventObject().GetTopLevelParent()
        hist = GlobalHistory().get_history(window)
        if hist is None:
            dprint1('!!! window does not have history!!!', window)
            return
        hist.make_entry(actions, menu_name='edit ' + name)

    def get_selected_page_text(self):
        i = self.nb.GetSelection()
        return self.nb.GetPageText(i)

    def set_selected_page_by_name(self, name):
        i = 0
        while i < self.nb.GetPageCount():
            txt = self.nb.GetPageText(i)
            if str(txt) == str(name):
                self.nb.SetSelection(i)
                return
            i = i+1

    def get_page_texts(self):
        return [str(self.nb.GetPageText(i))
                for i in range(self.nb.GetPageCount())]

    def adjust_elp(self, artist):
        pass

    def onNBChanged(self, evt):
        tw = wx.GetApp().TopWindow
        if tw.appearanceconfig.setting['generate_more_refresh']:
            self.GetTopLevelParent().deffered_force_layout()
        evt.Skip()


class artist_image_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_image_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_image as fig_image
        obj = fig_image.FigImage([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_spec_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_spec_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_spec as fig_spec
        obj = fig_spec.FigSpec([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_contour_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_contour_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_contour as fig_contour
        obj = fig_contour.FigContour([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()
        self.target_type = 0

    def Enable(self, value):
        #       print 'enabling contour widget'
        wx.Panel.Enable(self, value)
        for elp in self.elp:
            elp.Enable(value)
        if not value:
            return
        if self.target_artist is None:
            return
        if self.target_artist() is None:
            return
        self.adjust_elp(self.target_artist())

    def set_value(self, artist):
        if type(artist.figobj) != self.target_figobj_type:
            print("wrong target")
            return
        self.adjust_elp(artist)
        return artist_panel.set_value(self, artist)

    def adjust_elp(self, artist):
        from matplotlib.collections import PathCollection
        from matplotlib.collections import LineCollection

        if isinstance(artist, PathCollection):
            self.elp[-2].Enable(True)
            self.elp[-1].Enable(False)
        else:
            self.elp[-2].Enable(False)
            self.elp[-1].Enable(True)


class artist_box_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_box_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_box as fig_box
        obj = fig_box.FigBox()
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_circle_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_circle_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_circle as fig_circle
        obj = fig_circle.FigCircle()
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_curve_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_curve_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_curve as fig_curve
        obj = fig_curve.FigCurve()
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_spline_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_spline_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_spline as fig_spline
        obj = fig_spline.FigSpline([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_surface_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_surface_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        from ifigure.mto.fig_surface import FigSurface
        obj = FigSurface([0], [0], [0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_trisurface_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_trisurface_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        from ifigure.mto.fig_trisurface import FigTrisurface
        obj = FigTrisurface([0], [0], [0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_axline_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_axline_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_axline as fig_axline
        obj = fig_axline.FigAxline([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_axlinec_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_axlinec_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_axlinec as fig_axlinec
        obj = fig_axlinec.FigAxlineC([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_axspan_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_axspan_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_axspan as fig_axspan
        obj = fig_axspan.FigAxspan([0, 1])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_axspanc_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_axspanc_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_axspanc as fig_axspanc
        obj = fig_axspanc.FigAxspanC([0, 1])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_tripcolor_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_tripcolor_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_tripcolor as fig_tripcolor
        obj = fig_tripcolor.FigTripcolor([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_triplot_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_triplot_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_triplot as fig_triplot
        obj = fig_triplot.FigTriplot([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_scatter_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_scatter_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_scatter as fig_scatter
        obj = fig_scatter.FigScatter([0], [0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_timetrace_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_timetrace_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        from ifigure.mto.fig_plot import TimeTrace
        obj = TimeTrace([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_stepplot_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_stepplot_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        from ifigure.mto.fig_plot import StepPlot
        obj = StepPlot([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_plot_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_plot_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_plot as fig_plot
        obj = fig_plot.FigPlot([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()

    def set_value(self, artist):
        if type(artist.figobj) != self.target_figobj_type:
            return
        self.elp[0].Enable(True)
        self.elp[1].Enable(True)
        if artist.figobj._mpl_cmd == 'errorbar':
            self.elp[2].Enable(True)
        if artist.figobj._mpl_cmd == 'plot':
            self.elp[2].Enable(False)
#       self.elp[0].IsEnabled()

        from ifigure.mto.fig_plot import CZLineCollection
        if artist.figobj.getvar('cz'):
            self.Freeze()
            self.elp[0].Enable(True)  # set false of first two item
            self.elp[0].Enable([False, False])  # set false of first two item
            self.elp[1].Enable(False)
            self.Thaw()
        super(artist_plot_widget, self).set_value(artist)


class artist_plotc_widget(artist_plot_widget):
    def __init__(self, parent, *args):
        super(artist_plot_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_plotc as fig_plotc
        obj = fig_plot.FigPlotC([0, 0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_text_widget(artist_panel,  base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_text_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_text as fig_text
        obj = fig_text.FigText('')
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_hist_widget(artist_panel,  base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_hist_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_hist as fig_hist
        obj = fig_hist.FigHist(np.arange(100))
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_eps_widget(artist_panel,  base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_eps_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_eps as fig_eps
        obj = fig_eps.FigEPS()
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_quiver_widget(artist_panel,  base_artist_widget):
    def __init__(self, parent, *args):
        import numpy as np
        super(artist_quiver_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_quiver as fig_quiver
        obj = fig_quiver.FigQuiver(np.array([[1, 1], [1, 1]]),
                                   np.array([[1, 1], [1, 1]]))
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_quiver3d_widget(artist_panel,  base_artist_widget):
    def __init__(self, parent, *args):
        import numpy as np
        super(artist_quiver3d_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_quiver as fig_quiver
        obj = fig_quiver.FigQuiver3D(np.array([[1, 1], [1, 1]]),
                                     np.array([[1, 1], [1, 1]]),
                                     np.array([[1, 1], [1, 1]]),
                                     np.array([[1, 1], [1, 1]]),
                                     np.array([[1, 1], [1, 1]]),
                                     np.array([[1, 1], [1, 1]]))
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_legend_widget(artist_panel,  base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_legend_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        obj = FigLegend()
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_legendtext_widget(artist_panel,  base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_legendtext_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_legendtext as fig_legendtext
        obj = fig_legendtext.FigLegendText()
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_arrow_widget(artist_panel,  base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_arrow_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_arrow as fig_arrow
        obj = fig_arrow.FigArrow()
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_arrange_widget(wx.Panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_arrange_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        list = [["", 'this is fake arrage palette', 1],
                ["text", '', 0],
                ["color", 'red', 0]]
        box = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(box)
        self.elp = EditListPanel(self, list)
        box.Add(self.elp, 1, wx.EXPAND)


class artist_axes_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, obj=None, *args):
        super(artist_axes_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_axes as fig_axes
        if obj is None:
            obj = fig_axes.FigAxes()
        ret = obj.property_in_palette_axes()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_insetaxes_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_insetaxes_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_axes as fig_axes
        obj = fig_axes.FigInsetAxes()
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_colorbar_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_colorbar_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_axes as fig_axes
        obj = fig_axes.FigColorBar()
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_colorbar_axes_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_colorbar_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_axes as fig_axes
        obj = fig_axes.FigColorBar()
        ret = obj.property_in_palette_axes()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_mdssource_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_mdssource_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mdsplus.fig_mds as fig_mds
        obj = fig_mds.FigMds()
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_fill_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_fill_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_fill as fig_fill
        obj = fig_fill.FigFill([0])
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_solid_widget(artist_panel, base_artist_widget):
    def __init__(self, parent, *args):
        super(artist_solid_widget, self).__init__(parent, *args)
        base_artist_widget.__init__(self)
        import ifigure.mto.fig_solid as fig_solid
        obj = fig_solid.FigSolid(np.zeros((3, 3, 3)))
        ret = obj.property_in_palette()
        self.tab, self.list = self.build_editlist_list(obj, ret)
        self.make_panel()


class artist_widgets(wx.Panel):
    def __init__(self, parent, *args):
        super(artist_widgets, self).__init__(parent, *args)

        vbox = wx.BoxSizer(wx.HORIZONTAL)
        #vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        self.panels = {}
        for key in self.panels:
            self.panels[key].Hide()

#       self.panels[mode].Show()
#       self.GetSizer().Add(self.panels[mode], 1, wx.EXPAND)
        self.mode = ''
        self.artist = None
        self.Bind(EDITLIST_CHANGED, self.onEL_Changed)
        self.Enable(False)

    def build_panels(self, plist):
        for name, cls in plist:
            self.panels[name] = cls(self)

    def switch_panel(self, mode):

        if mode not in self.panels:
            return False
        if self.mode == mode:
            return True
        if self.mode != '':
            self.panels[self.mode].Hide()
            self.GetSizer().Detach(self.panels[self.mode])

        try:
            '''just in case'''
            self.GetSizer().Detach(self.panels[mode])
            self.panels[mode].Hide()
        except:
            print('cannnot remove')
        self.GetSizer().Add(self.panels[mode], 1, wx.EXPAND)
        for p in self.panels:
            self.panels[p].Hide()
        self.panels[mode].Show()
#       self.panels[mode].Layout()
        self.mode = mode

        FindFrame(self).Layout()
        return True

    #   self.Fit()
    def enable(self, val=True):
        self.Enable(val)
        if self.mode != '':
            self.panels[self.mode].Enable(val)

    def onTD_Replace(self, evt):
        for mode in self.panels:
            if self.panels[mode].target_artist is None:
                continue
            if self.panels[mode].target_artist() is None:
                continue
            self.panels[mode].replace_target_artist(evt)


class panel1(artist_widgets):
    plist = [('text', artist_text_widget),
             #               ('legendtext', artist_legendtext_widget),
             ('legend', artist_legend_widget),
             ('image', artist_image_widget),
             ('spec', artist_spec_widget),
             ('contour', artist_contour_widget),
             ('tricontour', artist_contour_widget),
             ('axline', artist_axline_widget),
             ('axlinec', artist_axlinec_widget),
             ('axspan', artist_axspan_widget),
             ('axspanc', artist_axspanc_widget),
             ('spline', artist_spline_widget),
             ('insetaxes', artist_insetaxes_widget),
             ('colorbar', artist_colorbar_widget),
             ('curve', artist_curve_widget),
             ('arrow', artist_arrow_widget),
             ('box', artist_box_widget),
             ('circle', artist_circle_widget),
             #               ('arrange', artist_arrange_widget),
             #               ('mdsplus', artist_mdssource_widget),
             ('plot', artist_plot_widget),
             ('plotc', artist_plot_widget),
             ('timetrace', artist_timetrace_widget),
             ('stepplot', artist_stepplot_widget),
             ('fill', artist_fill_widget),
             ('surface', artist_surface_widget),
             ('revolve', artist_surface_widget),
             ('trisurface', artist_trisurface_widget),
             ('tripcolor', artist_tripcolor_widget),
             ('triplot', artist_triplot_widget),
             ('eps', artist_eps_widget),
             ('quiver', artist_quiver_widget),
             ('quiver3d', artist_quiver3d_widget),
             ('hist', artist_hist_widget),
             ('scatter', artist_scatter_widget),
             ('solid', artist_solid_widget), ]

    try:
        import ifigure.mdsplus
        plist.append(('mdsplus', artist_mdssource_widget))
    except:
        pass

    plistd = {}
    for name, cls in plist:
        plistd[name] = cls

    def __init__(self, parent, quick, *args):
        super(panel1, self).__init__(parent, *args)

        if not quick:
            self.build_panels(panel1.plist)
        else:
            self.build_panels([panel1.plist[0]])
        self.artists = None

    def append_panel(self, mode):
        self.panels[mode] = panel1.plistd[mode](self)
        self.panels[mode].Show()
        self.panels[mode].Layout()

    def update_panel(self, mode=None):
        focus = self.FindFocus()
        if mode is None:
            mode = self.mode
        if self.artists is None:
            self.enable(False)
            return
        if len(self.artists) == 0:
            self.enable(False)
            return
        if self.artists[0]() is None:
            self.enable(False)
            return
        if self.artists[0]().figobj is None:
            self.artists = None
            self.enable(False)
            return
#        print self.artists[0]().figobj.get_figpage()
#        print self.GetTopLevelParent().canvas._figure.figobj
        if (self.artists[0]().figobj.get_figpage() is not
                self.GetTopLevelParent().canvas._figure.figobj):
            self.enable(False)
            return
        if mode in self.panels:
            if self.mode == mode:
                self.enable(True)
            self.panels[mode].set_value(self.artists[0]())
            self.panels[mode].set_target_artist(self.artists[0]())
            self.panels[mode].set_target_artist_mul(self.artists)
        if focus is not None:
            focus.SetFocus()

    def change_artist_panel(self, figobj):
        name = figobj.get_namebase()

        if name not in self.panels:
            if name in panel1.plistd:
                self.append_panel(name)
            else:
                self.enable(False)
                return

        self.switch_panel(name)
        self.enable(True)
        self.update_panel(name)

    def onTD_Selection(self, evt):
        if len(evt.selections) == 1:
            sel = evt.selections[0]()
            if sel is None:
                return
            if sel.figobj is None:
                return
            self.artists = evt.selections
            self.change_artist_panel(sel.figobj)

        elif len(evt.selections) == 0:
            # no selection -> disable
            self.enable(False)
        else:
            # multiple selection
            #  only enabled when all artists has same type of figobj
            #  and in the same axes
            self.enable(False)
            if any([obj() is None for obj in evt.selections]):
                return
            if any([obj().figobj is None for obj in evt.selections]):
                return

            figobjs = [obj().figobj for obj in evt.selections]
            parents = [obj().figobj.get_parent() for obj in evt.selections]

            if any([type(figobjs[0]) != type(obj) for obj in figobjs]):
                return
            if any([parents[0] is not p for p in parents]):
                return

            sel = evt.selections[0]()
            self.artists = evt.selections
            self.change_artist_panel(sel.figobj)

            self.enable(True)

    def onEL_Changed(self, evt):
        tw = wx.GetApp().TopWindow
        if tw.appearanceconfig.setting['generate_more_refresh']:
            wx.CallAfter(self.Update)

        actions, name = evt.artist_panel.set_artist_property(evt)
        if actions is None:
            return
        evt.artist_panel.add_action_to_history(evt, actions, name)
#       self.GetTopLevelParent().canvas.draw()


class panel2(artist_widgets):
    def __init__(self, parent, *args, **kwargs):
        super(panel2, self).__init__(parent, *args, **kwargs)

        wd = artist_axes_widget(self)
        mode = 'ax'+'.'.join(wd.tab)
        self.panels = {mode: wd}
        self.mode = ''
        self.ax = None
        self._ax_tab = 'common'
        self._cb_tab = 'common'
        self._target_3d = False

    def mode_name(self, figobj):
        tab, item = figobj.property_in_palette_axes()
        if isinstance(figobj, FigColorBar):
            mode = 'cb'+'.'.join(tab)
        else:
            mode = 'ax'+'.'.join(tab)
        return mode

    def append_new_panel(self, figaxes):
        if self.mode in self.panels:
            self._ax_tab = self.panels[self.mode].get_selected_page_text()
        for p in self.panels:
            self.panels[p].Hide()
            self.GetSizer().Detach(self.panels[p])
        mode = self.mode_name(figaxes)
        self.panels[mode] = artist_axes_widget(self, obj=figaxes)
        self.panels[mode].set_selected_page_by_name(self._ax_tab)
        self.panels[mode].Enable(True)

    def update_panel(self):
        focus = self.FindFocus()
        if self.ax is None:
            return
        if self.ax() is None:
            self.enable(False)
            return
        if self.ax().figobj is None:
            self.ax = None
            self.enable(False)
            return

        self._ax_tab = self.panels[self.mode].get_selected_page_text()
        mode_name = self.mode_name(self.ax().figobj)

        if self.mode != mode_name:
            if not mode_name in self.panels:
                self.append_new_panel(self.ax().figobj)
            self.switch_panel(mode_name)
            self.Layout()

        self.panels[mode_name].set_target_artist(self.ax())
        self.panels[mode_name].set_value(self.ax())

        pos = [elp.GetViewStart() for elp in self.panels[mode_name].elp]
        self.panels[mode_name].set_selected_page_by_name(self._ax_tab)
        for p, elp in zip(pos, self.panels[mode_name].elp):
            elp.Scroll(p)
        self.enable(True)
        if focus is not None:
            focus.SetFocus()

    def onTD_Selection(self, evt):
        if not hasattr(evt.GetEventObject(), "axes_selection"):
            return
        axes = evt.GetEventObject().axes_selection
        disable = False
        if axes is None:
            disable = True
        elif axes() is None:
            disable = True
        elif not hasattr(axes(), "figobj"):
            disable = True
        elif axes().figobj is None:
            disable = True

        if disable:
            self.enable(False)
            return
        self.enable(True)

        if isinstance(evt.GetTreeDict(), FigPage):
            return
        if len(evt.GetTreeDict()._artists) == 0:
            self.ax = None
            self.enable(False)
            return

        ax = evt.GetTreeDict()._artists[0].axes
        if ax is None:
            self.ax = None
            self.enable(False)
        else:
            self.ax = weakref.ref(ax)
            self.enable(True)
            if self.mode in self.panels:
                self._ax_tab = self.panels[
                    self.mode].get_selected_page_text()

            self.update_panel()

    def set_axes(self, ax):
        if ax is None:
            self.ax = None
            self.enable(False)
        else:
            self.ax = weakref.ref(ax)
            self.enable(True)

    def onEL_Changed(self, evt):
        tw = wx.GetApp().TopWindow
        if tw.appearanceconfig.setting['generate_more_refresh']:
            wx.CallAfter(self.Update)

        actions, name = evt.artist_panel.set_artist_property(evt)
        if actions is None:
            return
        if hasattr(evt, 'signal'):
            if evt.signal == 'need_adjustscale':
                actions.append(UndoRedoFigobjMethod(self.ax(),
                                                    'adjustrange', None))
        evt.artist_panel.add_action_to_history(evt, actions, name)
        self.GetTopLevelParent().canvas.draw()
        if self.mode in self.panels:
            self._ax_tab = self.panels[
                self.mode].get_selected_page_text()
        # do I need this????
        # self.update_panel()



from __future__ import print_function
from ifigure.mto.treedict import TreeDict
import ifigure
import os
from ifigure.utils.cbook import isInMainThread, LoadImageFile
import shlex
import subprocess
import threading
import traceback
import wx
import time

from ifigure.utils.get_username import get_username
usr = get_username()

sltime = 0.25  # sleep time during wait


def run_ssh_wait_and_retry(args, kargs, verbose=False):
    retry = 5
    ktry = 0
    while (ktry < retry):
        try:
            # if verbose:
            #    p = subprocess.Popen(args,
            #                         universal_newlines=True,
            #                         **kargs)
            # else:
            p = subprocess.Popen(args, stderr=subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 universal_newlines=True,
                                 **kargs)
        except:
            print(traceback.format_exc())

        while p.poll() == None:
            #if isInMainThread(): wx.Yield()
            time.sleep(sltime)
        stat = p.wait()

        err = p.stderr.readlines()
        success = True
        for x in err:
            if x.find('reset connection') != -1:
                print(x)
                success = False
            elif x.find('lost connection') != -1:
                print(x)
                success = False
            elif len(x) > 0:
                print(x)
                # success=False
            else:
                pass
        if success:
            break

        ktry = ktry+1
        print("retrying communicaiton : (" + str(ktry) + ")")
    return p


def run_ssh_no_retry(args, kargs, verbose=False):
    try:
        if verbose:
            p = subprocess.Popen(args,
                                 universal_newlines=True,
                                 **kargs)
        else:
            p = subprocess.Popen(args, stderr=subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 universal_newlines=True,
                                 **kargs)
    except:
        print(traceback.format_exc())

    return p


class PyConnection(TreeDict):
    def __init__(self, *args, **kargs):
        super(PyConnection, self).__init__(*args, **kargs)
        self.setvar('server', 'cmodws60')
        self.setvar('port', 22)
        self.setvar('user', usr)
        self.setvar('use_ssh', True)
        self.setvar('verbose', True)
        self.setvar('queue_type', 'sbatch')
        self.setvar('nocheck', False)
        self.setvar('multiplex', True)
        self.setvar('persist', 600)

    @classmethod
    def isPyConnection(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'connection'

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = LoadImageFile(path, 'world_link.png')
        return [idx1]

    def get_multiplex_opts(self):
        multiplex, persist = self.getvar('multiplex', 'persist')

        if multiplex:
            opts = ["-oControlMaster=auto",
                    "-oControlPersist="+str(persist),
                    "-oControlPath=~/.ssh/cm-%r@%h:%p-"+str(os.getpid())]
        else:
            opts = ["-oControlMaster=no", ]
        return opts

    def get_auth_opts(self, no_password=False):
        auth = self.getvar('auth_type')
        if no_password:
            opts = " -oPasswordAuthentication=no -oPreferredAuthentications=" + auth + " "
        else:
            opts = " -oPreferredAuthentications=" + auth + " "
        return opts

    def tree_viewer_menu(self):
        # return MenuString, Handler, MenuImage
        menu = [('Setting...', self.onSetting, None),
                ('---', None, None), ]
        return menu + super(PyConnection, self).tree_viewer_menu()

    def setSetting(self, value):
        self.setvar('use_ssh', value[1][0])
        self.setvar('server', str(value[1][1][0]))
        self.setvar('port', int(value[1][1][1]))
        self.setvar('user', str(value[1][1][2]))
        self.setvar('queue_type', str(value[2]))
        self.setvar('multiplex', bool(value[3]))
        self.setvar('persist', int(value[4]))
        self.setvar('auth_type', value[5])
        self.setvar('nocheck', not bool(value[6]))

    def onSetting(self, evt=None):
        server, port, user, use_ssh, queue_type, multiplex, persist, nocheck, auth_type = self.getvar('server', 'port',
                                                                                                      'user',
                                                                                                      'use_ssh',
                                                                                                      'queue_type',
                                                                                                      'multiplex',
                                                                                                      'persist',
                                                                                                      'nocheck',
                                                                                                      'auth_type')
        if use_ssh is None:
            use_ssh = True
        if server is None:
            server = 'cmodws60.psfc.mit.edu'
        if port is None:
            port = 22
        if user is None:
            user = usr  # os.getnev('USER')
        if queue_type is None:
            queue_type = 'sbatch'  # os.getnev('USER')
        if nocheck is None:
            nocheck = False
        if multiplex is None:
            multiplex = True
        if persist is None:
            persist = 600
        if auth_type is None:
            auth_type = "publickey"

        s_queue = {"style": wx.CB_DROPDOWN,
                   "choices": ["sbatch", "qsub"]}
        auth_choices = {"style": wx.CB_DROPDOWN,
                        "choices": ["publickey", "gssapi-with-mic"]}

        l2 = [["server", str(server), 0],
              ["port", str(port), 0],
              ["user", str(user), 0], ]

        l1 = [["", "SSH connection", 2],
              [None, (use_ssh, [server, str(port), user]), 27,
               ({'text': 'use SSH'}, {'elp': l2})],
              ["queue type", str(queue_type), 4, s_queue],
              [None, multiplex, 3, {"text": "use connection multiplexing"}],
              ["control persist (s)", int(persist), 400, {}],
              ["queue type", str(auth_type), 4, auth_choices],

              [None, not nocheck, 3, {"text": "connection check/retry"}], ]
        proj_tree_viewer = wx.GetApp().TopWindow.proj_tree_viewer
        proj_tree_viewer.OpenPanel(l1, self, 'setSetting',
                                   title='Connection setting')

    def PutFile(self, src, dest, nowait=False):
        '''
        put a file in a remote server
        '''
        server, port, user, use_ssh, nocheck = self.getvar('server', 'port',
                                                           'user', 'use_ssh', 'nocheck')

        if nocheck is None:
            nocheck = False

        if use_ssh:
            opts = ' '.join(self.get_multiplex_opts())
            command = 'scp ' + opts + ' -P '+str(port) + ' ' + src + \
                ' ' + user + '@' + server+':' + dest
            args = shlex.split(command)
            args = command
            kargs = {'shell': True}
        else:
            command = 'cp ' + src + ' ' + dest
            args = command
            kargs = {'shell': True}

        verbose = self.getvar('verbose')

        if verbose:
            print(command)
        if nowait:
            p = run_ssh_no_retry(args, kargs, verbose=verbose)
        elif nocheck:
            p = run_ssh_no_retry(args, kargs, verbose=verbose)
            p.wait()
        else:
            p = run_ssh_wait_and_retry(args, kargs, verbose=verbose)
        return p

    def GetFile(self, src, dest, nowait=False, nocheck=False):
        '''
        get a file in a remote server
        '''
        server, port, user, use_ssh, nocheck = self.getvar('server', 'port',
                                                           'user', 'use_ssh',
                                                           'nocheck')
        if nocheck is None:
            nocheck = False
        if use_ssh:
            opts = ' '.join(self.get_multiplex_opts())
            command = 'scp ' + opts + ' -P '+str(port) + ' ' + user + \
                '@'+server+':'+src + ' ' + dest
            args = shlex.split(command)
            kargs = {}
            args = command
            kargs = {'shell': True}
        else:
            command = 'cp ' + src + ' ' + dest
            args = command
            kargs = {"shell": True}

        verbose = self.getvar('verbose')
        if verbose:
            print(command)

        if nowait:
            p = run_ssh_no_retry(args, kargs, verbose=verbose)
        elif nocheck:
            p = run_ssh_no_retry(args, kargs, verbose=verbose)
            p.wait()
        else:
            p = run_ssh_wait_and_retry(args, kargs, verbose=verbose)
        return p

    def GetFiles(self, srcs, dest_dir):
        '''
        get multiple files to a directory
        '''
        server, port, user, use_ssh, nocheck = self.getvar('server', 'port',
                                                           'user', 'use_ssh',
                                                           'nocheck')
        if nocheck is None:
            nocheck = False
        src = r'\{' + ','.join(srcs) + r'\}'

        if use_ssh:
            opts = ' '.join(self.get_multiplex_opts())
            command = 'scp ' + opts + ' -P '+str(port) + ' ' + user + \
                '@'+server+':'+src + ' ' + dest_dir
            args = shlex.split(command)
            kargs = {}
            args = command
            kargs = {'shell': True}
        else:
            command = 'cp ' + src + ' ' + dest_sie
            args = command
            kargs = {"shell": True}

        verbose = self.getvar('verbose')
        if verbose:
            print(command)

        p = run_ssh_no_retry(args, kargs, verbose=verbose)
        p.wait()

    def Execute(self, command, nowait=False, nocheck=False, force_ssh=False):
        server, port, user, use_ssh, nocheck = self.getvar('server', 'port',
                                                           'user', 'use_ssh',
                                                           'nocheck')
        if nocheck is None:
            nocheck = False
        if use_ssh or force_ssh:
            opts = ' '.join(self.get_multiplex_opts())
            command = 'ssh ' + opts + ' -x -p '+str(port)+' ' + \
                user+'@'+server + " '" + command + "'"
            args = shlex.split(command)
            kargs = {}
            args = command
            kargs = {'shell': True}
        else:
            args = command
            kargs = {"shell": True}

        verbose = self.getvar('verbose')
        if verbose:
            print(command)

        if nowait:
            p = run_ssh_no_retry(args, kargs, verbose=verbose)
        elif nocheck:
            p = run_ssh_no_retry(args, kargs, verbose=verbose)
            p.wait()
        else:
            p = run_ssh_wait_and_retry(args, kargs, verbose=verbose)
        return p

        '''    
        if self.getvar('verbose'):
            print(command)

        try:
            p = subprocess.Popen(args, stderr=subprocess.STDOUT,
                                 stdout=subprocess.PIPE,
                                 universal_newlines=True,
                                 **kargs)
        except:
            print(traceback.format_exc())
        print('a new process (id =' + str(p.pid) + ') is running')
        if not nowait:
            while p.poll() == None:
                #if isInMainThread(): wx.Yield()
                time.sleep(sltime)
            print(p.wait())
        return p
        '''



from __future__ import print_function
from ifigure.widgets.simple_shell import ShellBase
import wx
import os
import wx.py.shell  # (wx4 removed this) wx.lib.shell
from wx.py.interpreter import Interpreter as IP

from ifigure.utils.wx3to4 import panel_SetToolTip
_wait = False
_test_shell = True


def exit_wait():
    globals()['_wait'] = False


def set_wait():
    globals()['_wait'] = True


def is_waiting():
    return _wait


panel = None


def check_debugger_instance():
    if panel is None:
        raise ValueError("debugger is not yet instanciated")
    else:
        panel.CheckDebuggerInstance()


class DebuggerInterpreter(IP):

    def __init__(self, *args, **kargs):
        IP.__init__(self, *args, **kargs)
        self._g_dict = {}
        self._l_dict = {}

    def runcode(self, code):
        """Execute a code object.

        When an exception occurs, self.showtraceback() is called to
        display a traceback.  All exceptions are caught except
        SystemExit, which is reraised.

        A note about KeyboardInterrupt: this exception may occur
        elsewhere in this code, and may not always be caught.  The
        caller should be prepared to deal with it.

        """
        try:
            exec(code, self._l_dict, self._g_dict)
        except SystemExit:
            raise
        except:
            self.showtraceback()
        else:
            from code import softspace
            import sys
            if softspace(sys.stdout, 0):
                print()


class DebugShell(ShellBase):
    def __init__(self, parent, id):
        self.lvar = {}
        super(DebugShell, self).__init__(
            parent, id, InterpClass=DebuggerInterpreter)
        self.clear_window()

    def clear_window(self):
        self.clear()
        self.prompt()

    def setBuiltinKeywords(self):
        '''
        in order not to overwrite quit/exit command
        '''
        pass

    def push(self, command, silent=False):
        frame = self.GetParent().frame

        if frame is not None:
            self.interp._g_dict = frame.f_globals
            self.interp._l_dict = frame.f_locals
            super(DebugShell, self).push(command, silent=silent)


class DebuggerPanel(wx.Panel):
    def __init__(self, parent, id, *args, **kargs):
        wx.Panel.__init__(self, parent, id,
                          style=wx.FRAME_FLOAT_ON_PARENT | wx.CLOSE_BOX)

        self.frame = None
        self._status = 'stop'
        self.SetSizer(wx.BoxSizer(wx.VERTICAL))
        msizer = wx.BoxSizer(wx.HORIZONTAL)
#       msizer = wx.BoxSizer(wx.HORIZONTAL)
#       msizer2 = wx.BoxSizer(wx.VERTICAL)
#       msizer2.Add(msizer, 1, wx.EXPAND)
        self.GetSizer().Add(msizer, 0, wx.EXPAND)
        d_labels = ['s', 'n', 'r', 'c', 'q']
        d_tips = ['step', 'next', 'return', 'continue',  'quit']
        d_buttons = [wx.Button(self, wx.ID_ANY, size=(40, 20)) for
                     button in d_labels]

        for b, l, t in zip(d_buttons, d_labels, d_tips):
            b.SetLabel(" "+l+" ")
            panel_SetToolTip(b, t)
            msizer.Add(b, 0, wx.ALL, 1)

            def call_debug_button(evt, obj=self, label=l):
                obj.onDebugButton(evt, label)
            self.Bind(wx.EVT_BUTTON, call_debug_button, b)

        from ifigure.utils.edit_list import TextCtrlCopyPasteGeneric

        if _test_shell:
            self.txt = DebugShell(self, wx.ID_ANY)
        else:
            self.txt = TextCtrlCopyPasteGeneric(self, wx.ID_ANY, '',
                                                style=wx.TE_PROCESS_ENTER)
        #self.Bind(wx.EVT_TEXT_ENTER, self.onTextEnter, self.txt)

        msizer2 = wx.BoxSizer(wx.HORIZONTAL)
        self.GetSizer().Add(msizer2, 1, wx.ALL | wx.EXPAND, 0)
        msizer2.Add(self.txt, 1, wx.ALL | wx.EXPAND, 0)
        self.st = wx.StaticText(self, wx.ID_ANY, ' '*50)
        self.st.Wrap(300)
        msizer.Add(self.st, 1, wx.ALL | wx.EXPAND, 1)

        self.d_buttons = d_buttons
        self.d = None
        self.Layout()
        self.Hide()
        self.Fit()
        globals()['panel'] = self
#       bdb.set_trace()

#   def Show(self):
#       for b in self.d_buttons:
#           b.Show()
#       wx.Panel.Show(self)

#   def Hide(self):
#       for b in self.d_buttons:
#           b.Hide()
#       wx.Panel.Hide(self)

    def onTextEnter(self, evt):
        if is_waiting():
            txt = str(self.txt.GetValue())
            if txt != '':
                try:
                    #                   print(eval(txt, self.frame.f_globals,
                    #                                 self.frame.f_locals))
                    exec(compile(txt, 'text', 'exec'), self.frame.f_globals, self.frame.f_locals)
                except:
                    print('evaluation failed')

    def onDebugButton(self, evt, l):
        if l == 's':  # step
            self.d.set_step()
            exit_wait()
        elif l == 'n':
            self.d.set_next(self.frame)
            exit_wait()
        elif l == 'c':
            self.d.set_continue()
            exit_wait()
        elif l == 'r':
            self.d.set_return(self.frame)
            exit_wait()
        elif l == 'p':
            from ifigure.widgets.dialog import textentry
            txt = str(self.txt.GetValue())
            if txt != '':
                try:
                    print(eval(txt, self.frame.f_globals,
                               self.frame.f_locals))
                except:
                    print('evaluation failed')
        elif l == 'q':
            self.StopDebugger()

    def StopDebugger(self):
        self.d.set_quit()
        exit_wait()
        self._status = 'stop'
        print('exiting debug mode')
        self.exit_debug_mode()

#   def CheckDebuggerInstance(self):
#       if self.d is None:
#           from ifigure.widgets.debugger_core import get_debugger
#           self.d = get_debugger(self)

    def Run(self, script_co, l1, l2, file, call_back):
        #       script_file = script.path2fullpath()
        #       file = self.d.canonic(script_file)
        #       print self.d
        if self.d is not None:
            return

        from ifigure.widgets.debugger_core import get_debugger
        self.d = get_debugger(self)
        self._status = 'running'
        self._file = file
        self._call_back = call_back
        self._call_count = 0
        self.target_code = None
        self.d.run(script_co, l1, l2)
        if self._status != 'stop':
            self.StopDebugger()

    def setStatus(self, mode, frame, *args):
        if self.target_code is None:
            self.target_code = frame.f_code
            self.target_frame = frame
        self.mode = mode
        self.args = args
        self.frame = frame

    def handle_user_line(self, frame):
        self.setStatus('line', frame)
#       if self.target_code.co_filename != frame.f_code.co_filename: return
        wx.GetApp().TopWindow.open_editor_panel()
        editor = self.GetParent().GetParent()
        if editor.get_current_file() != frame.f_code.co_filename:
            editor.GoDebugMode(frame.f_code.co_filename, enter=True)
        editor.ShowDebugCurrentLine(frame.f_code.co_filename,
                                    frame.f_lineno)

        self.st.SetLabel(os.path.basename(frame.f_code.co_filename) +
                         ', line#:' + str(frame.f_lineno))

        self.Layout()
        set_wait()
        while is_waiting():
            wx.Yield()

    def handle_user_return(self, frame, return_value):
        print(('return', frame, return_value, self._call_count))
        if self._call_count > 0:
            self._call_count = self._call_count - 1
        else:
            self._status = 'stop'
            print(('exiting debug mode : return value = ', return_value))
            self.exit_debug_mode()

    def handle_user_call(self, frame, return_value):
        self._call_count = self._call_count + 1

    def handle_user_exception(self, frame, exc_info):
        self._status = 'stop'
        print(('exiting debug mode : exception info = ', exc_info))
        self.exit_debug_mode()

    def exit_debug_mode(self):
        self._call_back()
        self.d = None

    @property
    def status(self):
        return self._status



import ifigure.utils.cbook as cbook
import ifigure.events
from ifigure.mto.fig_axes import FigAxes, FigInsetAxes
from ifigure.widgets.canvas.custom_picker import linehit_test, abs_d
from ifigure.widgets.undo_redo_history import GlobalHistory, UndoRedoArtistProperty, UndoRedoFigobjProperty, UndoRedoFigobjMethod
import weakref
import ifigure.utils.pickle_wrapper as pickle
import numpy as np

'''
 GenericPointHolder
 GenericPoint

 classes to handle (x. y) points which is expressed
 in various coordinate system.

 it provides...
     interface to change setting
     routine to convert coords

'''


class GenericPointsHolder(object):
    def __init__(self, num=0, **kywds):
        self._gp_points = []
        trans = self.getvar('trans')
        transaxes = self.getvar('transaxes')
        if num != 0:
            for k in range(num):
                self.add_gp(GenericPoint(0, 0))

    def set_parent(self, parent):
        self.set_gp_figpage()
        for gp in self._gp_points:
            if hasattr(gp, '_d_bk'):
                #                print self.get_figpage()._artists
                gp.dict_to_gp(gp._d_bk, self)

    def destroy(self):
        for p in self._gp_points:
            self.rm_gp(p)

    def add_gp(self, gp):
        self._gp_points.append(gp)
        return len(self._gp_points)-1

    def rm_gp(self, gp):
        self._gp_points.remove(gp)

    def get_gp(self, i):
        return self._gp_points[i]

    def num_gp(self):
        return len(self._gp_points)

    def set_gp_by_vars(self):
        trans = self.getvar('trans')
        transa = self.getvar('transaxes')
        for k, gp in enumerate(self._gp_points):
            transaxes = transa[k]
            if trans is not None:
                gp.set_trans(trans[k*2:(k+1)*2])
            if transaxes is None:
                ax = None
            elif transaxes == 'default':
                ax = self.get_figaxes()
            else:
                try:
                    ax = find_by_full_path(transaxes)
                except:
                    ax = None
            if ax is not None:
                self.set_gp_figaxes(gp, ax, no_conversion=True)

    def set_gp_figpage(self):
        figpage = self.get_figpage()
        if figpage is None:
            return
        for gp in self._gp_points:
            gp.set_figpage(figpage)
            if gp.trans is None:
                figaxes = self.get_figaxes()
                if figaxes is None:
                    gp.set_trans(['figure', 'figure'])
                else:
                    gp.set_trans(['axes', 'axes'])
                    gp.set_figaxes(figaxes)

    def set_gp_figaxes(self, p, figaxes, no_conversion=False):
        '''
        change figaxes used for transform of gp
        '''
        if figaxes is None:
            return
        if isinstance(p, GenericPoint):
            gp = p
        else:
            gp = self.get_gp(p)
        # print gp.get_figaxes(), gp.get_figaxes()
        if gp.get_figaxes() is not None:
            t_from = gp.get_gp_transform(dir='both')  # <-it should be here?
        else:
            no_conversion = True
        old_figaxes = gp.get_figaxes()
        if old_figaxes is not None:
            old_figaxes.remove_update_client(self)
        gp.set_figaxes(figaxes)
        figaxes.add_update_client(self)
        if len(figaxes._artists) == 0:
            return
        if no_conversion:
            return
        t_to = gp.get_gp_transform(dir='both')
#        t_from = gp.get_gp_transform(dir='both')
        gp.x, gp.y = gp.convert_trans_p((gp.x, gp.y), t_from, t_to)

    def move_gp_points(self, index, dx, dy, action=True):
        gp = self.get_gp(index)
        new_x, new_y = gp.calc_new_points(dx, dy)
        if action:
            action1 = UndoRedoFigobjMethod(self._artists[0],
                                           'gppoint'+str(index),
                                           (new_x, new_y))
            return action1
        else:
            m = getattr(self, 'set_'+name1)
            m((new_x, new_y))

    def get_device_point(self, index):
        return self._gp_points[index].get_device_point()

    def set_device_point(self, index, x, y):
        return self._gp_points[index].set_device_point(x, y)

    def set_gppoint0(self, xy, a):
        self.set_gp_point(0, xy[0], xy[1])
        self.set_update_artist_request()

    def set_gppoint1(self, xy, a):
        self.set_gp_point(1, xy[0], xy[1])
        self.set_update_artist_request()

    def set_gppoint2(self, xy, a):
        self.set_gp_point(2, xy[0], xy[1])
        self.set_update_artist_request()

    def set_gppoint3(self, xy, a):
        self.set_gp_point(3, xy[0], xy[1])
        self.set_update_artist_request()

    def get_gppoint0(self, a):
        return self.get_gp_point(0)

    def get_gppoint1(self, a):
        return self.get_gp_point(1)

    def get_gppoint2(self, a):
        return self.get_gp_point(2)

    def get_gppoint3(self, a):
        return self.get_gp_point(3)

    def get_gp_point(self, i):
        return self._gp_points[i].x, self._gp_points[i].y

    def set_gp_point(self, i, x, y, trans=None):
        if trans is not None:
            self._gp_points[i].trans = trans
        self._gp_points[i].x = x
        self._gp_points[i].y = y

    def change_gp_trans(self, gp, i0, value):
        '''
        change transform method of gp
        '''
        gp.change_trans(i0, value)

    def gp_hittest_p(self, evt, gp):
        x1, y1 = gp.get_device_point()
        return abs_d(x1, y1, evt.x, evt.y) < 9

    def gp_hittest_rect(self, evt, gp1, gp2):
        x1, y1 = gp1.get_device_point()
        x2, y2 = gp2.get_device_point()

        if abs_d(x1, y1, evt.x, evt.y) < 9:
            return 1
        if abs_d(x2, y1, evt.x, evt.y) < 9:
            return 2
        if abs_d(x2, y2, evt.x, evt.y) < 9:
            return 3
        if abs_d(x1, y2, evt.x, evt.y) < 9:
            return 4

#        print (linehit_test(evt.x, evt.y, x1, y1, x2, y2),
#            evt.x >= min([x1, x2]),
#            evt.y >= min([y1, y2]),
#            evt.x <= max([x1, x2]),
#            evt.y <= max([y1, y2]))

        if (linehit_test(evt.x, evt.y, x1, y1, x2, y2) and
            evt.x >= min([x1, x2])-2 and
            evt.y >= min([y1, y2])-2 and
            evt.x <= max([x1, x2])+2 and
                evt.y <= max([y1, y2])+2):
            hit = True
            return 5
        return 0

    def gp_hittest_line(self, evt, gp1, gp2):
        x1, y1 = gp1.get_device_point()
        x2, y2 = gp2.get_device_point()
        v1 = (x2 - x1, y2 - y1)
        v2 = (evt.x - x1, evt.y - y1)
        l1 = np.sqrt(v1[0]*v1[0]+v1[1]*v1[1])
        l2 = np.sqrt(v2[0]*v2[0]+v2[1]*v2[1])
        d = abs(v1[0]*v2[1] - v1[1]*v2[0])/l1     # distance
        d2 = (v1[0]*v2[0] + v1[1]*v2[1])/l1/l1  # ratio
        # print d, d2

        if (0 < d2 < 1) and d < 5:
            return 1
        return 0

    def set_gp_trans(self, value, a=None):
        i = value[0][0]
        idx = value[0][1]
        v = value[1]
        gp = self.get_gp(i)
        self.change_gp_trans(gp, idx, v)

    def get_gp_trans(self, a=None, extra=None):
        i = extra[0]
        idx = extra[1]
        gp = self.get_gp(i)
        return gp.trans[idx]

    def gp_canvas_menu(self, gp):
        return gp.make_canvas_menu(self)

    def gp_holder_canvas_menu(self):
        '''
        provide simple version of gp menu
        allow to change the setting of all gp points
        at once
        '''
        a0 = ['figure', 'axes', 'data']
        funcs = [None]*3
        for k in range(3):
            def func(evt, value=a0[k], holder=self):
                ac = []
                for i in range(holder.num_gp()):
                    action1 = UndoRedoFigobjMethod(self._artists[0],
                                                   'gp_trans', value)
                    action1.set_extrainfo((i, 0))
                    action2 = UndoRedoFigobjMethod(self._artists[0],
                                                   'gp_trans', value)
                    action2.set_extrainfo((i, 1))
                    ac.append(action1)
                    ac.append(action2)
#                    gp = holder.get_gp(i)
#                    holder.change_gp_trans(gp, 0, value)
#                    holder.change_gp_trans(gp, 1, value)
                window = evt.GetEventObject().GetTopLevelParent()
                hist = GlobalHistory().get_history(window)
                hist.make_entry(ac, menu_name='change trans')
            funcs[k] = func

        value = []
        for i in range(self.num_gp()):
            value = value + self.get_gp(i).trans
#        flag = True
#        for i in range(len(value)):
#            if value[i] != value[0]: flag = False
        m = ["*figure", "*axes", "*data"]
        if 'figure' in value:
            m[0] = '^figure'
        if 'axes' in value:
            m[1] = '^axes'
        if 'data' in value:
            m[2] = '^data'

        menu = [("+Coods...", None, None),
                (m[0], funcs[0], None),
                (m[1],  funcs[1], None),
                (m[2],  funcs[2], None),
                ("!",  None, None), ]
        if self.get_figaxes() is None:
            menu.extend([
                ("Select Axes...", self.onSelTransaxes_all, None), ])
        return menu

    def onSelTransaxes_all(self, evt):
        def onSelTransaxes_cb(figaxes, holder=self):
            for gp in self._gp_points:
                holder.onSelTransaxes_cb(figaxes, gp)

        if self.num_gp() == 0:
            return
        gp = self.get_gp(0)
        ifigure.events.SendInteractiveAxesSelection(self, w=evt.GetEventObject(),
                                                    callback=onSelTransaxes_cb,
                                                    figaxes=gp.get_figaxes(),
                                                    handler=evt.GetEventObject())

    def onSelTransaxes(self, evt, gp):
        def onSelTransaxes_cb(figaxes, gp0=gp):
            self.onSelTransaxes_cb(figaxes, gp0)
        ifigure.events.SendInteractiveAxesSelection(self,
                                                    w=evt.GetEventObject(),
                                                    callback=onSelTransaxes_cb,
                                                    figaxes=gp.get_figaxes(),
                                                    handler=evt.GetEventObject())

    def onSelTransaxes_cb(self, figaxes, gp):
        #        print figaxes.get_full_path()
        self.set_gp_figaxes(gp, figaxes)

    def save_data(self, fid=None):
        data = []
        for gp in self._gp_points:
            data.append(gp.gp_to_dict())
        pickle.dump(data, fid)

    def save_data2(self, data):
        #        import traceback
        #        traceback.print_stack()
        try:
            # if target axes was destroyed already
            # this step fails
            # this happens when closing window,, and so on???
            if hasattr(self, "_loaded_gp_data"):
                d = self._loaded_gp_data
            else:
                d = [gp.gp_to_dict() for gp in self._gp_points]
        except:
            d = []
        data['GenericPointsHolder'] = (1, d)
        return data

    def load_data(self, fid=None):
        h2 = pickle.load(fid)
        if self.hasp("loaded_property"):
            lp = self.getp("loaded_property")
        else:
            lp = []
        lp.append(h2)
        self.setp("loaded_property", lp)

    def load_data2(self, data):
        if not 'GenericPointsHolder' in data:
            return
        h2 = data['GenericPointsHolder'][1]
        if len(h2) > len(self._gp_points):
            for k in range(len(h2) - len(self._gp_points)):
                self.add_gp(GenericPoint(0, 0))
        elif len(h2) < len(self._gp_points):
            for k in range(-len(h2) + len(self._gp_points)):
                self._gp_points.remove(self._gp_points[-1])
#        for k, gp in enumerate(self._gp_points):
#            gp.dict_to_gp(h2[k], self)

        self._loaded_gp_data = h2


class GenericPoint(object):
    def __init__(self, x=0, y=0, trans=['figure']*2,
                 figaxes=None, figpage=None):
        object.__init__(self)
        self.x = x
        self.y = y
        self.trans = trans  # 'figure', 'axes', 'points'

        self.set_figpage(figpage)
        self.set_figaxes(figaxes)

    def __repr__(self):
        return 'GenericPoint('+str(self.x) + ',' + str(self.y) + ':'+','.join(self.trans)+')'

    def gp_to_dict(self):
        if self.figaxes() is not None:
            xy = self.figaxes().get_rect()
            ax = ((xy[0]+xy[2])/2, (xy[1]+xy[3])/2)
        else:
            ax = None
        d = {"x": self.x,
             "y": self.y,
             "trans": self.trans,
             "ax": ax}
        return d

    def dict_to_gp(self, d, holder):
        self.x = d["x"]
        self.y = d["y"]
        self.trans = d["trans"]

        if d["ax"] is None:
            self.set_figaxes(None)
        else:
            rect = []
            if holder.get_figpage() is None:
                # if parent is not yet given
                # this should be postponed.
                # till next set_parent is called
                self._d_bk = d
                return

            for td in holder.get_figpage().walk_tree():
                if isinstance(td, FigAxes):
                    rect.append((td, td.get_rect()))

            # set reference axes to the closest axes
            area = [abs((xy[1][2]-xy[1][0])*(xy[1][3]-xy[1][1]))
                    for xy in rect]

            idx = [x[0] for x in sorted(enumerate(area), key=lambda x:x[1])]
            dist = [np.sqrt((d["ax"][0] - (rect[i][1][0] + rect[i][1][2])/2.)**2 +
                            (d["ax"][1] - (rect[i][1][1] + rect[i][1][3])/2.)**2) for i in idx]

            i = np.argmin(dist)

            holder.set_gp_figaxes(self, rect[idx[i]][0])
            self.x = d["x"]
            self.y = d["y"]

            if hasattr(self, '_d_bk'):
                del self._d_bk
        return self

    def set_figaxes(self, figaxes=None):
        #        print 'setting figaxes', figaxes
        #        import traceback
        #        traceback.print_stack()
        if figaxes is not None:
            self.figaxes = weakref.ref(figaxes, self.lost_figaxes)
        else:
            self.figaxes = cbook.WeakNone()

    def get_figaxes(self):
        return self.figaxes()

    def set_figpage(self, figpage=None):
        if figpage is not None:
            self.figpage = weakref.ref(figpage, self.lost_figaxes)
        else:
            self.figpage = cbook.WeakNone()

    def get_figpage(self):
        return self.figpage()

    def change_trans(self, i0, value):
        '''
        change transform method of gp
        '''
        trans = [name for name in self.trans]
        trans[i0] = value
        t_from = self.get_gp_transform(dir='both')
        t_to = self.get_gp_transform(dir='both', name=trans)
        self.x, self.y = self.convert_trans_p((self.x, self.y), t_from, t_to)
        self.trans = trans

    def change_trans_figaxes(self, gp, figaxes):
        '''
        change figaxes used for transform of gp
        '''
        t_from = gp.get_gp_transform(dir='both')
        self.set_figaxes(figaxes)  # ?
        t_to = gp.get_gp_transform(dir='both')
        gp.x, gp.y = gp.convert_trans_p((gp.x, gp.y), t_from, t_to)

    def lost_figaxes(self, ref):
        pass
        # print 'referred axes is deleted'

    def set_trans(self, value):
        self.trans = value

    def get_gp_transform(self, dir='both', name=None):
        ''' 
        get transform of genericpoint
        '''
        def get_transform(figaxes, figpage, n):
            if figaxes is None:
                t = figpage._artists[0].transFigure
            elif n == 'data':
                t = figaxes._artists[0].transData
            elif n == 'axes':
                t = figaxes._artists[0].transAxes
            elif n == 'figure':
                t = figpage._artists[0].transFigure
            return t

        ans = []
        if dir == 'x' or dir == 'both':
            if name is None:
                n = self.trans[0]
            else:
                try:
                    n = name[0]
                except Exception:
                    n = name
            ans.append(get_transform(self.figaxes(), self.figpage(), n))
        if dir == 'y' or dir == 'both':
            if name is None:
                n = self.trans[1]
            else:
                try:
                    n = name[1]
                except Exception:
                    n = name
            ans.append(get_transform(self.figaxes(), self.figpage(), n))
        return ans

    def make_canvas_menu(self, holder):
        def make_menu_str(i):
            trans = self.trans[i]
            a0 = ['figure', 'axes', 'data']
            a = ['figure', 'axes', 'data']
            funcs = [None]*3
            for k in range(3):
                if a[k] == trans:
                    a[k] = '^'+a[k]
                else:
                    a[k] = '*'+a[k]

                def func(evt, value=a0[k], i0=i, h=holder, gp=self):
                    id = h._gp_points.index(gp)
                    action1 = UndoRedoFigobjMethod(h._artists[0],
                                                   'gp_trans', value)
                    action1.set_extrainfo((id, i0))
                    window = evt.GetEventObject().GetTopLevelParent()
                    hist = GlobalHistory().get_history(window)
                    hist.make_entry([action1], menu_name='change trans')

                funcs[k] = func
            return a, funcs

        def make_menu_list(dir, a, funcs):
            m1 = [("+"+dir+" Coords",  None, None),
                  (a[0],  funcs[0], None),
                  (a[1],  funcs[1], None),
                  (a[2],  funcs[2], None),
                  ("!",  None, None)]
            return m1

        a, funcs = make_menu_str(0)
        m1 = make_menu_list('X', a, funcs)
        a, funcs = make_menu_str(1)
        m2 = make_menu_list('Y', a, funcs)
        m = m1 + m2
        if self.figpage() is not None:
            def onSelTransaxes(evt, obj=holder, gp=self):
                obj.onSelTransaxes(evt, gp)
            if holder.get_figaxes() is None:
                m.extend([("Select Axes...", onSelTransaxes, None), ])
#            m = m + [("Select Axes...",  onSelTransaxes, None)]
        return m

    def get_device_point(self):
        t = self.get_gp_transform(dir='both')
        x1, y1 = self.convert_trans_p((self.x, self.y), t, None)
        return int(x1), int(y1)

    def calc_new_points(self, dx, dy):
        '''
        calculate new x, y for transport of dx, dy
        '''
        x_old = self.x
        y_old = self.y
        x1, y1 = self.get_device_point()
        self.set_device_point(x1+dx, y1+dy)
        x_new = self.x
        y_new = self.y
        self.x = x_old
        self.y = y_old
        return x_new, y_new

    def set_device_point(self, x1, y1):
        t = self.get_gp_transform(dir='both')
        self.x, self.y = self.convert_trans_p((x1, y1), None, t)

    def convert_trans_rect(self, r, t1=None, t2=None):
        n1 = self.convert_trans_p([r[0], r[1]], t1, t2)
        n2 = self.convert_trans_p([r[0]+r[2], r[1]+r[2]], t1, t2)

        w = n2[0][0]-n1[0][0]
        h = n2[0][1]-n1[0][1]
        rect = [n1[0][0], n1[0][1], w, h]
        return rect

    def convert_trans_p(self, p, t1=None, t2=None):
        '''
        it dose 
        1) transfomr to device using t1
        2) transform to new coord by t2.inverted()
        '''
        n1 = np.array([p[0], p[1]]).reshape(1, 2)
        if t1 is not None:
            n1_new1 = t1[0].transform(n1)
            if t1[0] != t1[1]:
                n1_new2 = t1[1].transform(n1)
                n1 = np.array([n1_new1[0][0], n1_new2[0][1]]).reshape(1, 2)
            else:
                n1 = n1_new1
        if t2 is not None:
            n1_new1 = t2[0].inverted().transform(n1)
            if t2[0] != t2[1]:
                n1_new2 = t2[1].inverted().transform(n1)
                n1 = np.array([n1_new1[0][0], n1_new2[0][1]]).reshape(1, 2)
            else:
                n1 = n1_new1
        return n1[0][0], n1[0][1]



from __future__ import print_function

import matplotlib
try:
    # matplotlib3 has this option
    matplotlib.set_loglevel("error")
except:
    pass

import ifigure
import matplotlib.cm
import weakref
import shutil
from ifigure.widgets.dlg_preference import PrefComponent
import tempfile
from ifigure.utils.cbook import register_idl_colormaps
from ifigure.utils.pid_exists import pid_exists
import sys
import site
from ifigure.utils.get_username import get_username
from matplotlib.lines import Line2D
from matplotlib.path import Path
from matplotlib.collections import PathCollection
import matplotlib
import re
import os
from os.path import expanduser
import traceback
from distutils.version import LooseVersion

from matplotlib.artist import ArtistInspector

from ifigure.utils.setting_parser import iFigureSettingParser as SP
import ifigure.utils.debug as debug

dprint1, dprint2, dprint3 = debug.init_dprints('ifiure_config')


isMPL2 = LooseVersion(matplotlib.__version__) >= LooseVersion("2.0")
isMPL33 = LooseVersion(matplotlib.__version__) >= LooseVersion("3.3")

if isMPL2:
    # let's use classic for now.
    import matplotlib.pyplot
    matplotlib.pyplot.style.use('classic')

pickle_protocol = 5

def artist_property_checker(obj, prop, values=None):
    #print('inspecting', obj, prop)
    if values is None:
        values = ArtistInspector(obj).get_valid_values(prop)
    matches = re.findall(r"\'(.+?)\'", values)
    #print('inspector returns ', values)
    #print('checking ...', matches)
    setter = getattr(obj, 'set_'+prop)
    getter = getattr(obj, 'get_'+prop)

    values = []
    matches_ans = []
    for i, p in enumerate(matches):
        try:
            setter(p)
            x = getter()
            values.append(x)
            matches_ans.append(p)
        except:
            dprint2(str(obj)+'do not know how to set ' +
                    prop + ' to ' + p.__repr__())
    #print('object returns...', values)
    return matches_ans, values


ifiguredir = ifigure.__path__[0]
#resourcedir= os.path.join(os.path.dirname(os.path.dirname(ifigure.__path__[0])), 'resources')
resourcedir = os.path.join(ifigure.__path__[0], 'resources')
icondir = os.path.join(resourcedir, 'icon')
home = expanduser("~")
usr = get_username()

rcdir = os.getenv("PISCOPERC", default="")
if rcdir == '':
    rcdir = os.path.join(home, '.ifigure_rc')

# piscope's own package site
# made in .ifigure_rc
# this directory is used to store package/modules used from
# simulation modules

site.USER_SITE = os.path.join(rcdir, '.local', 'site-packages')
site.USER_BASE = os.path.join(rcdir, '.local')
if not os.path.exists(rcdir):
    os.mkdir(rcdir)
if not os.path.exists(site.USER_BASE):
    os.mkdir(site.USER_BASE)
if not os.path.exists(site.USER_SITE):
    os.mkdir(site.USER_SITE)
sys.path.insert(0, site.USER_SITE)
file = os.path.join(site.USER_SITE, '__init__.py')
if not os.path.exists(file):
    fid = open(file, 'w')
    fid.close()
file = os.path.join(site.USER_SITE, 'piscopelib')
if not os.path.exists(file):
    os.mkdir(file)
    file = os.path.join(file, '__init__.py')
    fid = open(file, 'w')
    fid.close()

geom_file = os.path.join(rcdir, 'gui_geom')

usr_template_dir = os.path.join(rcdir, 'template')
if not os.path.exists(usr_template_dir):
    os.mkdir(usr_template_dir)
usr_script_template_dir = os.path.join(usr_template_dir, 'script')
if not os.path.exists(usr_script_template_dir):
    os.mkdir(usr_script_template_dir)

p = SP()
s = p.read_setting('pref.general_config')
if s['root_work_directory'] == '\'default\'':
    #    tempdir_base=os.path.join(tempfile.gettempdir(), 'piscope_'+os.getenv('USER'))
    tempdir_base = os.path.join(tempfile.gettempdir(), 'piscope_'+usr)
    if not os.path.exists(tempdir_base):
        os.makedirs(tempdir_base)
#    tempdir_base=os.path.join(tempdir_base, os.getenv('USER'))
#    if not os.path.exists(tempdir_base):
#        os.mkdir(tempdir_base)
else:
    tempdir_base = os.path.expanduser(s['root_work_directory'])
    tempdir_base = os.path.join(tempdir_base, 'piscope_'+usr)
    if not os.path.exists(tempdir_base):
        os.makedirs(tempdir_base)


crashed_process = []


#
# temporary directory is <hostname>.PID_xxxx
#    we add hostname to the temporary file,
#    in order to avoid deleteing the work directory
#    when the tmporary directory is on the share drive
#    and when a user opens the same project file from
#    more than two different computers
#
#    hostname is taken by using socket.gethostname()
#    if it fails
#    we use a nominal hostname = "HOST"
#
for item0 in os.listdir(tempdir_base):
    item = item0.split('.')[-1]
    if item.startswith('pid'):
        pid = int(item[3:])
        if not pid_exists(pid):
            try:
                dname = os.path.join(tempdir_base, item0)
                for item2 in os.listdir(dname):
                    if item2.startswith('###untitled_') or item2.startswith('.###ifigure_'):
                        f = os.path.join(tempdir_base, item0, item2)
                        shutil.rmtree(f)
                if len(os.listdir(dname)) == 0:
                    try:
                        print(('removing past crush dir', dname))
                        shutil.rmtree(dname)
                    except:
                        print('Failed to remove crush dir')
            except:
                pass

try:
    import socket
    name = socket.gethostname().split('.')[0]
except:
    name = "HOST"
tempdir = os.path.join(tempdir_base, name + '.pid'+str(os.getpid()))

if os.path.exists(tempdir) == False:
    os.mkdir(tempdir)


def tempdir_clean(obj):
    if os.path.exists(tempdir):
        print(('removing tempdir ', tempdir))
        shutil.rmtree(tempdir)


scratch = os.path.join(rcdir, 'ifigure_scratch')
canvas_scratch = os.path.join(rcdir, 'canvas_scratch')
if os.path.exists(canvas_scratch) == False:
    os.mkdir(canvas_scratch)
canvas_scratch_page = os.path.join(rcdir, 'canvas_scratch_page')
if os.path.exists(canvas_scratch_page) == False:
    os.mkdir(canvas_scratch_page)
canvas_scratch_axes = os.path.join(rcdir, 'canvas_scratch_axes')
if os.path.exists(canvas_scratch_axes) == False:
    os.mkdir(canvas_scratch_axes)
vv_scratch = os.path.join(rcdir, 'vv_scratch')
st_scratch = os.path.join(rcdir, 'ifigure_subtree')
pick_r = 10

print('reading extra color maps')
register_idl_colormaps()

#
#   palette data
#
collist = ['blue', 'g', 'red',
           'cyan', 'magenta', 'yellow',
           'black', 'white', 'none', 'grey', 'darkgrey', 'lightgrey']
# collist=['b', 'g', 'r',
#         'c', 'm', 'y',
#         'k', 'w', 'none', 'grey', 'darkgrey' , 'lightgrey']
collistf = ['blue', 'green', 'red',
            'cyan', 'magenta', 'yellow',
            'black', 'white', 'grey', 'darkgrey', 'lightgrey']
markerlist = ['None', '.', ',', 'o', 'v', '^', '<', '>',
              '1', '2', '3', '4', 's', 'p', '*', 'h', 'H', '+', 'x', 'D', 'd',
              '|', '_']
markernames = ['None', 'pint', 'pixel', 'o', 'triangle_down',
               'triangle_up', 'triangle_left', 'triangle_right',
               '1', '2', '3', '4', 's', 'p', '*', 'h', 'H', 'plus', 'x', 'D', 'd',
               'vline', 'hline']

linestylelist = ['None', '-', '--', '-.', ':']
linestylenames = ['None', '-', '--', '-.', 'colon']

linewidthlist = ['1.0', '1.5', '2.0', '2.5', '3.0']

arrowstylename = ["Curve", "CurveB", "BracketB",
                  "CurveFilledB", "CurveA", "CurveAB", "CurveFilledA",
                  "CurveFilledAB", "BracketA", "BracketAB", "Fancy",
                  "Simple", "Wedge", "BarAB"]
arrowstylelist = ["-", "->", "-[", "-|>", "<-", "<->",
                  "<|-", "<|-|>", "]-", "]-[", "fancy",
                  "simple", "wedge", "|-|"]

colormap_from_reference =  [('Perceptually Uniform Sequential', [
            'viridis', 'plasma', 'inferno', 'magma', 'cividis']),
         ('Sequential', [
            'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
            'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
            'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn']),
         ('Sequential (2)', [
            'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
            'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia',
            'hot', 'afmhot', 'gist_heat', 'copper']),
         ('Diverging', [
            'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
            'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic']),
         ('Cyclic', ['twilight', 'twilight_shifted', 'hsv']),
         ('Qualitative', [
            'Pastel1', 'Pastel2', 'Paired', 'Accent',
            'Dark2', 'Set1', 'Set2', 'Set3',
            'tab10', 'tab20', 'tab20b', 'tab20c']),
         ('Miscellaneous', [
            'flag', 'prism', 'ocean', 'gist_earth', 'terrain', 'gist_stern',
            'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg',
            'gist_rainbow', 'rainbow', 'jet', 'nipy_spectral', 'gist_ncar'])]

colormaplist = sum([x[1] for x in colormap_from_reference],[])

#
#  Artist may not reture the same value as given by setter...
#  Also, matplotlib has a routine to get a list of valid property values.
#  With get_valid_values, it may be possible to implement more automatic
#  property list generation.
#
#  The following is an initial attempt. Needs to check more how well
#  matplotlib is organized in this aspect.

obj = PathCollection(Path.unit_rectangle())
#plinestylelist, plinestyle_rlist = artist_property_checker(obj, 'linestyle')
pedgecolorlist, pedgecolor_rlist = artist_property_checker(
    obj, 'edgecolor', values="'"+"','".join(collist)+"'")
plinestylelist = ['solid', 'dotted', 'dashed', 'dashdot']
plinestyle_rlist = ['solid', 'dotted', 'dashed', 'dashdot']
if isMPL2:
    plinestylelist = plinestylelist[:4]
    plinestyle_rlist = plinestyle_rlist[:4]


def colormap_list():
    return colormaplist


def scratch_file():
    return scratch


def color_list():
    return collist


def color_list_face():
    return collistf


def marker_list():
    return markerlist


def linestyle_list():
    return linestylelist


def linewidth_list():
    return linewidthlist


def plinestyle_list():
    return plinestylelist


def arrowstyle_list():
    return [(arrowstylename[i], arrowstylelist[i]) for i in range(len(arrowstylename))]


class iFigureConfig(PrefComponent):
    setting = None

    def __init__(self):
        PrefComponent.__init__(self, 'General')
        if iFigureConfig.setting is None:
            p = SP()
            iFigureConfig.setting = p.read_setting('pref.general_config')

    def save_setting(self):
        p = SP()
        p.write_setting('pref.general_config', iFigureConfig.setting)

    def get_dialoglist(self):

        s = iFigureConfig.setting
        if s["delbook_on_windowclose"] == 1:
            text = 'no'
        else:
            text = 'yes'
        list = [["Model add-on path", s['usr_addon_dir'], 0, None],
                ["Root work dir. (need restart)", s['root_work_directory'], 200, None],
                ["WikiHelp front", s['wikihelp_front'], 0, None],
                ["WikiHelp base", s['wikihelp_base'], 0, None],
                ["Default keep book in proj", text, 1,
                 {"values": ["yes", "no"]}]]
        hint = ["path to scripts/modules used in models",
                'root directory of temporary work files. ("default": tempfile.gettempdir()/os.getenv(\'USER\'))',
                "wiki front page",
                "wiki base page",
                "Default delete book data when window is closed"]

        return list, hint

    def set_dialog_result(self, value):
        s = iFigureConfig.setting
        if str(value[4]) == 'yes':
            val = 1
        else:
            val = 0
        s["usr_addon_dir"] = str(value[0])
        s['root_work_directory'] = str(value[1])
        s["wikihelp_front"] = str(value[2])
        s["wikihelp_base"] = str(value[3])
        s["delbook_on_windowclose"] = not (val)



from __future__ import print_function
#
#   MDSScope no-window
#
#       no window version of scope.
#       a user can load data without opening book
#       on a screen. Internally it uses exact same
#       ScopeEngine to process MDSplus sessions
#
#   example:
#     >> o =scopenw(proj.lh_scope)
#     >> o.SetShot(1101119004)
#     >> o.LoadShot(verbose = False)
#     >> o.close()
#     >> from ifigure.widgets.book_viewer import BookViewer
#     >> proj.lh_scope.Open(BookViewer)
#
# *******************************************
#     Copyright(c) 2012- S.Shiraiwa
# *******************************************
__author__ = "Syun'ichi Shiraiwa"
__copyright__ = "Copyright, S. Shiraiwa, PiScope Project"
__credits__ = ["Syun'ichi Shiraiwa"]
__version__ = "1.0"
__maintainer__ = "Syun'ichi Shiraiwa"
__email__ = "shiraiwa@psfc.mit.edu"
__status__ = "beta"

import wx
import sys
import time
import weakref
import logging
import threading
from six.moves import queue as Queue
import os
import shutil
import numpy
import traceback
import collections
import multiprocessing as mp
#import wx.aui as aui
import ifigure
import ifigure.events
import ifigure.utils.cbook as cbook

#from ifigure.widgets.canvas.ifigure_canvas import ifigure_canvas
from ifigure.widgets.property_editor import property_editor
from ifigure.widgets.panel_checkbox import PanelCheckbox
from ifigure.widgets.statusbar import StatusBar, StatusBarSimple
import ifigure.widgets.dialog as dialog

#from ifigure.mto.fig_book import FigBook
#from ifigure.mto.fig_page import FigPage
#from ifigure.mto.fig_axes import FigAxes
from ifigure.mdsplus.fig_mds import FigMds, MDSJobSet
from ifigure.mdsplus.fig_mdsdata import FigMdsData
from ifigure.mto.figobj_param import FigobjParam, FigobjData
from ifigure.widgets.book_viewer import BookViewerFrame

import ifigure.mdsplus.mdsplusr as remote_mds

from ifigure.utils.setting_parser import iFigureSettingParser as SettingParser
from ifigure.mdsplus.mds_sp_worker_pool import MDSSPWorkerPool
from ifigure.mdsplus.mds_mpth_worker_pool import MDSMPWorkerPool, MDSTHWorkerPool
#from ifigure.mdsplus.fig_mds import build_init_local_global_session
from ifigure.mdsplus.fig_mds import build_current_shot_session
from ifigure.mdsplus.fig_mds import add_connect_session
from ifigure.mdsplus.event_listener import MDSSCOPE_EVT_MDSEVENT
from ifigure.mdsplus.mdsscope import ScopeEngine, message, set_call_after_queue

#
#  debug setting
#
import ifigure.utils.debug as debug
debug.debug_default_level = 1
dprint1, dprint2, dprint3 = debug.init_dprints('MDSScope')


class FakeTextCtrl(object):
    def __init__(self, *args, **kargs):
        self._value = ''

    def SetValue(self, value):
        self._value = value

    def GetValue(self):
        return self._value


class FakeBookViewerFrame(object):
    def __init__(self, *args, **kargs):
        self._title = ''
        self._parent = ''
        self._status_txt = ['']*10
        self._print_status = True
        if "book" in kargs:
            self.book = kargs["book"]
            del kargs["book"]
  #          self.book.set_open(True)
        else:
            self.book = None

        self.ipage = 0

    def SetTitle(self, title):
        self._title = title
        print(('title: ', title))

    def GetTitle(self):
        return self._title

    def SetStatusText(self, txt, idx):
        self._status_txt[idx] = txt
        if self._print_status:
            print(('status: ', txt))

    def get_page(self, ipage=None):
        if ipage is None:
            f_page = self.book.get_page(self.ipage)
        else:
            f_page = self.book.get_page(ipage)
        return f_page


class MDSScopeNW(FakeBookViewerFrame,  ScopeEngine):
    def __init__(self, *args, **kargs):
        if "show_prop" in kargs:
            show_prop = kargs["show_prop"]
            del kargs["show_prop"]
        else:
            show_prop = False
        if "worker" in kargs:
            self.workers = kargs['worker']
            del kargs["worker"]
        else:
            self.workers = None

        super(MDSScopeNW, self).__init__(*args, **kargs)
        ScopeEngine.__init__(self, no_window=True)

        if self.book is not None:
            #            for page in self.book.walk_page():
            self.prepare_dwglobal(self.book)
            if not self.book.hasvar('mdsplus_server'):
                p = SettingParser()
                v = p.read_setting('mdsplus.default_mdsserver')
                self.book.setvar('mdsplus_server',
                                 v['server'])
#                                 'direct::CMOD')
            if not self.book.hasvar('mdsscript_main'):
                self.book.setvar('mdsscript_main',
                                 True)
            if not self.book.hasvar('mdsscope_autogrid'):
                self.book.setvar('mdsscope_autogrid',
                                 True)
            if not self.book.hasvar('mdsscope_listenevent'):
                self.book.setvar('mdsscope_listenevent',
                                 True)
            if not self.book.hasvar('dwscope_filename'):
                self.book.setvar('dwscope_filename', '')
            from ifigure.mdsplus.fig_mdsbook import convert2figmdsbook
            convert2figmdsbook(self.book)

        self.g = {}  # global variabls for mainthread scripting
        self._title_mdsjob_count = 0
        self._mdsjob_count = 0
        self._cur_shot = 0
        self._mds_exit_status = 0
        self._figmds_list = []
        self.mpanalysis = False
        self._mpanalysis_mode = True
        self.parallel_page = True
        self.parallel_shot = True
        self.parallel_sec = True
        self.init_beginning = True
        self.init_page = False
        self.init_shot = False
        self.init_sec = False
        self.debug_mode = False
        self.timer = None

        self.previous_shot_set = [[]]
        self.event_dict = {}
        self.InitUI()
        from numpy import linspace
        self._shot_dict = {'linspace': linspace}
#        self.Thaw()

        proj = self.book.get_root_parent()
        proj = self.book.get_root_parent()
        if self.workers is None:
            if proj.setting.has_child('mdsplus_worker'):
                self.workers = proj.setting.mdsplus_worker
            else:
                file = os.path.join(ifigure.__path__[0], 'add_on',
                                    'setting', 'module', 'mdsplus_worker.py')

                workers = proj.setting.add_absmodule(file)
                workers.rename('mdsplus_worker')
                self.workers = workers

        p = SettingParser()
        p.set_rule('global_set', {}, nocheck=True)
        self.scope_setting = p.read_setting('mdsplus.scope_setting')

        from ifigure.ifigure_config import rcdir, ifiguredir
        mname = 'mdsplus.fig_mds_startup_script'
        def_file = os.path.join(*([ifiguredir] + mname.split('.')))
        user_file = os.path.join(*([rcdir] + mname.split('.')))+'.py'
        if not os.path.exists(user_file):
            shutil.copy(def_file, user_file)
        self.startup_script = user_file
        dc = {}
        dg = {}
        from ifigure.mdsplus.fig_mds import read_scriptfile
        exec(read_scriptfile(self.startup_script), dg, dc)
        self.startup_values = dc

        self._start_mds_threads()  # start session runner and event listener

        self.workers.call_method('onStartWorker')
#        self.start_pool()
        self.open_engine()

    def __del__(self):
        self.close_engine()

    def close(self):
        self.close_engine()

    def InitUI(self):
        self.txt_shot = FakeTextCtrl()

    def show_page(self, ipage=0, last=False, first=False):
        self.ipage = ipage
        self.prepare_dwglobal(self.book)

    def _handle_apply_abort(self, allshot=True, figaxes='all',
                            do_apply=False):
        from ifigure.mdsplus.mdsscope import mds_thread
        #mds_thread = globals()['mds_thread']
        mds_thread._time = time.time()
        if self._mode == 'apply' or do_apply:
            self._mode = 'abort'
            try:
                self.eval_mdsdata(allshot, figaxes=figaxes)
            except:
                dprint1('Error happend while evaluating mdsplus data')
                dprint1(traceback.format_exc())
                self.eval_mdsdata_done(status=-1)
        else:
            c = message('abort', (self,))
            mds_thread.queue.put(c)

    def show_mdsjob_count(self):
        if (self._title_mdsjob_count ==
                self._mdsjob_count):
            return

        if self._mdsjob_count != 0:
            txt = ' --'+'{:3d}'.format(self._mdsjob_count) + '--'
        else:
            txt = ''
        work_on_title = False
        if work_on_title:
            title = self.GetTitle()
            if self._title_mdsjob_count != 0:
                title = title[:-8]
            title = title + txt
            self.SetTitle(title)
        self._title_mdsjob_count = self._mdsjob_count
        self.SetStatusText(txt, 0)

    def SetShot(self, number):
        '''
        Set Shot Number
        '''
        self.txt_shot.SetValue(str(number))

    def LoadData(self, blocking=True, allshot=True, figaxes='all',
                 do_apply=True, verbose=True):
        '''
        LoadData() 
        LoadData(False) : no blocking mode
        '''
        o_printstatus = self._print_status
        self._print_status = verbose
        if blocking:
            m = Queue.Queue()
            set_call_after_queue(m)
        self._handle_apply_abort(allshot=allshot,
                                 figaxes=figaxes,
                                 do_apply=do_apply)
        if not blocking:
            return
        while True:
            v = m.get(True)
            callable, args, kargs = v
            callable(*args, **kargs)
            if callable == self.eval_mdsdata_done:
                break
        set_call_after_queue(None)
        self._print_status = o_printstatus
        return



from __future__ import print_function

__author__ = "Syun'ichi Shiraiwa"
__copyright__ = "Copyright, S. Shiraiwa, PiScope Project"
__credits__ = ["Syun'ichi Shiraiwa"]
__version__ = "0.5"
__maintainer__ = "S. Shiraiwa"
__email__ = "shiraiwa@psfc.mit.edu"
__status__ = "beta"


from ifigure.mto.fig_obj import FigObj
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import ifigure.utils.cbook as cbook
import ifigure.widgets.canvas.custom_picker as cpicker
import numpy as np
import weakref
import logging
import matplotlib
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty, UndoRedoFigobjProperty, UndoRedoFigobjMethod


class FigAnnotate(FigObj):
    def __init__(self, s, xy, xytext=None, xycoords='data',
                 textcoords='data', arrowprops=None, **kywds):

        if xytext == None:
            xytext = xy
        if arrowprops == None:
            arrowprops = dict(arrowstyle="->",
                              connectionstyle="arc3")
        self._objs = []  # for debug....
        if "draggable" in kywds:
            self.setvar("draggable", kywds["draggable"])
            del kywds["draggable"]
        else:
            self.setvar("draggable", True)
        args = ()
        super(FigAnnotate, self).__init__(*args, **kywds)

        self.setvar("xy", xy)
        self.setvar("s",  s)
        self.setvar("xytext", xytext)
        self.setvar("xycoords", xycoords)
        self.setvar("textcoords", textcoords)
        self.setvar("arrowprops", arrowprops)
        self.setp('use_var', True)

    @classmethod
    def isFigAnnotate(self):
        return True

    @classmethod
    def can_have_child(self, child=None):
        return False

    @classmethod
    def get_namebase(self):
        return 'annotate'

    @classmethod
    def property_in_file(self):
        return []

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return ["use_var", "xy", "xytext", "s"]

    @classmethod
    def property_in_palette(self):
        return ["tab1", "tab2"], [["color"], ["color"]]

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'arrow.bmp')
        return [idx1]

    def isDraggable(self):
        return self._var["draggable"]

    def generate_artist(self):
        container = self.get_container()
        if self.isempty() is False:
            return

        lp = self.getp("loaded_property")

        if self.getp('use_var'):

            xy = self.eval("xy")
            s = self.eval("s")
            xytext = self.eval("xytext")
            xycoords = self.eval("xycoords")
            textcoords = self.eval("textcoords")
            arrowprops = self.eval("arrowprops")
            self.setp("s", s)
            self.setp("xy", xy)
            self.setp("xytext", xytext)
            self.setp("textcoords", xycoords)
            self.setp("xycoords", xycoords)
            self.setp("arrowprops", arrowprops)
            self.setp('use_var', False)
        else:
            s = self.getp("s")
            xy = self.getp("xy")
            xytext = self.getp("xytext")
            xycoords = self.getp("xycoords")
            textcoords = self.getp("textcoords")
            arrowprops = self.getp("arrowprops")

        self._artists = [container.annotate(s, xy,
                                            xytext=xytext,
                                            xycoords=xycoords,
                                            textcoords=textcoords,
                                            arrowprops=arrowprops)]
        for artist in self._artists:
            artist.figobj = self
            artist.figobj_hl = []

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = artist

        val = []
        for a in self._artists:
            val.append(self.get_artist_property(a))
        if len(val) != 0:
            self.setp("loaded_property", val)

        if len(artistlist) != 0:
            self.highlight_artist(False, artistlist)
            container = self.get_container()
            for a in artistlist:
                #            a is axes in this case
                a.remove()

        super(FigAnnotate, self).del_artist(artistlist)

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist
        container = self._artists[0].get_figure()
        w = container.get_figwidth()*container.get_dpi()
        h = container.get_figheight()*container.get_dpi()
        if val == True:
            for a in alist:

                arr = a.arrow_patch.get_verts()
                hl = matplotlib.lines.Line2D(arr[:, 0]/w, arr[:, 1]/h,
                                             marker='s',
                                             color='k', linestyle='None',
                                             markerfacecolor='None',
                                             figure=container,
                                             transform=container.transFigure,)
                container.lines.extend([hl])
                a.figobj_hl.append(hl)
        else:
            for a in alist:
                for hl in a.figobj_hl:
                    hl.remove()
                    a.figobj_hl = []

    def get_artist_extent(self, a):
        arr = a.arrow_patch.get_verts()
        x = arr[:, 0]
        y = arr[:, 1]
        return [min(x), max(x), min(y), max(y)]

    def picker_a0(self, a, evt):
        from ifigure.widgets.canvas.custom_picker import linehit_test, abs_d
        hit, extra, type, loc = super(FigAnnotate, self).picker_a0(a, evt)

        arr = a.arrow_patch.get_verts()
        x = arr[:, 0]
        y = arr[:, 1]
        self._drag_mode = 0

        hit = False
        if abs_d(x[0], y[0], evt.x, evt.y) < 5:
            self._drag_mode = 1
            hit = True
        elif abs_d(x[-1], y[-1], evt.x, evt.y) < 5:
            self._drag_mode = 2
            hit = True
        for i in range(len(x)-1):
            if (linehit_test(evt.x, evt.y, x[i], y[i], x[i+1], y[i+1]) < 5 and
                evt.x > min([x[i], x[i+1]]) and
                evt.y > min([y[i], y[i+1]]) and
                evt.x < max([x[i], x[i+1]]) and
                    evt.y < max([y[i], y[i+1]])):
                hit = True
                self._drag_mode = 3
                break
        return hit, extra, type, loc

    def drag_a(self, a, evt, shift=None):
        print(self._picker_a_type)
        shift = evt.guiEvent.ShiftDown()
        redraw = super(FigAnnotate, self).drag_a(a, evt, shift=shift)
        return redraw

    def dragdone_a(self, a, evt, shift=None):
        shift = evt.guiEvent_memory.ShiftDown()
        redraw = super(FigAnnotate, self).dragdone_a(a, evt, shift=shift)
        dx = evt.x - self._st_p[0]
        dy = evt.y - self._st_p[1]

        xy = self.getp("xy")
        xytext = self.getp("xytext")
        xycoords = self.getp("xycoords")
        textcoords = self.getp("textcoords")

        hist = self.get_root_parent().app.history
        hist.start_record()

        if self._drag_mode & 2 != 0:
            t1 = self.coordsname2transform(self._parent._artists[0],
                                           xycoords)
            xyd = self.convert_transform_point(t1, None, xy)
            xyd = (xyd[0]+dx, xyd[1]+dy)
            xy = self.convert_transform_point(None, t1, xyd)
            action1 = UndoRedoFigobjProperty(self._artists[0],
                                             'xy', xy)
            hist.add_history(action1)
        if self._drag_mode & 1 != 0:
            t2 = self.coordsname2transform(self._parent._artists[0],
                                           textcoords)
            xytextd = self.convert_transform_point(t2, None, xytext)
            xytextd = (xytextd[0]+dx, xytextd[1]+dy)
            xytext = self.convert_transform_point(None, t2, xytextd)
            action2 = UndoRedoFigobjProperty(self._artists[0],
                                             'xytext', xytext)
            hist.add_history(action2)

        hist.stop_record()



import sys
import socket
import shlex
import threading
import os
import binascii
import subprocess
import time
import tempfile
from six.moves import socketserver
import traceback
import ifigure.utils.pickle_wrapper as pickle
from ifigure.utils.pickled_pipe import PickledPipe
l = 3000000


class JobRunner(object):
    def __init__(self, host, port):
        self.p = None
        self.host = 'localhost'
        self.port = port

    def __del__(self):
        from signal import SIGTERM, SIGKILL
        if self.p is not None:
            os.kill(self.p.pid, SIGKILL)

    def terminate(self):
        if self.p is not None:
            self.p.kill()
        self.p == None

    def run(self, jobset):
        import numpy as np
        return {'x': np.arange(l), 'y': np.arange(l), 'error message': ''}



import wx
from ifigure.widgets.book_viewer import BookViewer
from ifigure.mto.fig_plot import FigPlot


class SliceViewer(BookViewer):
    def __init__(self, *args, **kargs):
        kargs['style'] = (wx.FRAME_FLOAT_ON_PARENT | wx.MINIMIZE_BOX
                          | wx.MAXIMIZE_BOX | wx.RESIZE_BORDER
                          | wx.SYSTEM_MENU | wx.CAPTION | wx.CLOSE_BOX
                          | wx.CLIP_CHILDREN)
#        print kargs
        BookViewer.__init__(self, *args, **kargs)
        f_page = self.get_page(0)
        f_page.set_section(2)
        f_axes = self.get_axes(0, 0)
        for name, child in f_axes.get_children():
            child.destroy()
        obj = FigPlot([0], [0])
        f_axes.add_child('xslice_plot1', obj)
        obj.set_suppress(True)
        obj = FigPlot([0], [0])
        f_axes.add_child('xslice_plot2', obj)
        obj.set_suppress(True)
        f_axes.realize()
        f_axes = self.get_axes(0, 1)
        for name, child in f_axes.get_children():
            child.destroy()
        obj = FigPlot([0], [0])
        f_axes.add_child('xslice_plot1', obj)
        obj.set_suppress(True)
        obj = FigPlot([0], [0])
        f_axes.add_child('xslice_plot2', obj)
        obj.set_suppress(True)
        f_axes.realize()

    def update_curve(self, idx, x, y, data1, data2):
        from ifigure.interactive import plot, nsec, isec, update, title, xlabel, ylabel
        from ifigure import _cursor_config as cconfig

        self.Raise()
        c1 = cconfig["1dcolor1"]
        c2 = cconfig["1dcolor2"]
        if idx == 0:
            c = c1
        else:
            c = c2

        f_axes = self.get_axes(0, 0)
        f_axes.set_bmp_update(False)
        title = 'x slice : x = '+str(x)
        self._update_curve(idx, f_axes, c, x, y, data1, title)
        f_axes = self.get_axes(0, 1)
        f_axes.set_bmp_update(False)
        title = 'y slice : y = '+str(y)
        self._update_curve(idx, f_axes, c, x, y, data2, title)
        self.draw()

    def _update_curve(self, idx, f_axes, c, x, y, data, title):

        if f_axes.num_child() == 0:
            obj = FigPlot([0], [0])
            f_axes.add_child('xslice_plot1', obj)
            obj = FigPlot([0], [0])
            f_axes.add_child('xslice_plot2', obj)
            f_axes.realize()
        j = 0
        for name, child in f_axes.get_children():
            if (isinstance(child, FigPlot)):
                if j == idx:
                    child.setvar('x', data[0])
                    child.setvar('y', data[1])
                    child.set_suppress(False)
                    child.reset_artist()
                    a = child._artists[0]
                    a.set_color(c)
                    break
                else:
                    j = j+1
        f_axes.adjust_axes_range()
        v = f_axes.get_title(f_axes._artists[0])
        v[0] = title
        v = f_axes.set_title(v, f_axes._artists[0])
        f_axes.set_bmp_update(False)



import os
import wx
from ifigure.utils.wx3to4 import GridSizer


class FindPanel(wx.Panel):
    def __init__(self, parent, id, *args, **kargs):
        wx.Panel.__init__(self, parent, id,
                          style=wx.FRAME_FLOAT_ON_PARENT | wx.CLOSE_BOX)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(hsizer)

        self.btn_cl = wx.Button(self, wx.ID_CLOSE, 'x', style=wx.BU_EXACTFIT)
        sizer0 = wx.BoxSizer(wx.VERTICAL)

        hsizer.Add(sizer0, 1, wx.ALL | wx.ALIGN_CENTER_VERTICAL)
        hsizer.Add(self.btn_cl, 0, wx.ALL | wx.ALIGN_TOP)

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        sizer0.Add(sizer, 1, wx.EXPAND | wx.ALL, 1)
        sizer0.Add(sizer2, 1, wx.EXPAND | wx.ALL, 1)
        label = wx.StaticText(self)
        label.SetLabel('Find : ')
        label2 = wx.StaticText(self)
        label2.SetLabel('Replace : ')
        from ifigure.utils.edit_list import TextCtrlCopyPasteGeneric
        self.txt = TextCtrlCopyPasteGeneric(self, wx.ID_ANY, '',
                                            style=wx.TE_PROCESS_ENTER)
        self.txt2 = TextCtrlCopyPasteGeneric(self, wx.ID_ANY, '',
                                             style=wx.TE_PROCESS_ENTER)
        self.btn_bw = wx.Button(self, wx.ID_ANY, 'Backward')
        self.btn_fw = wx.Button(self, wx.ID_ANY, 'Forward')
        gsizer = GridSizer(1, 2)
        gsizer.Add(self.btn_bw, wx.ALL | wx.ALIGN_CENTER_VERTICAL)
        gsizer.Add(self.btn_fw, wx.ALL | wx.ALIGN_CENTER_VERTICAL)

        #from ifigure.ifigure_config import icondir
        #imageFile =os.path.join(icondir, '16x16', 'close.png')
        # bitmap=wx.Bitmap(imageFile)
        #self.btn_cl = wx.BitmapButton(self, bitmap=bitmap)

        self.Bind(wx.EVT_BUTTON, parent.onHitFW, self.btn_fw)
        self.Bind(wx.EVT_BUTTON, parent.onHitBW, self.btn_bw)
        self.Bind(wx.EVT_BUTTON, parent.onHitCL, self.btn_cl)
        self.Bind(wx.EVT_TEXT_ENTER, parent.onRunFind, self.txt)

        self.btn_replace = wx.Button(self, wx.ID_ANY, 'Replace')
        self.btn_replaceall = wx.Button(self, wx.ID_ANY, 'Replace All')
        self.Bind(wx.EVT_BUTTON, parent.onReplace, self.btn_replace)
        self.Bind(wx.EVT_BUTTON, parent.onReplaceAll, self.btn_replaceall)

        sizer.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL)
        sizer.Add(self.txt, 1, wx.ALL | wx.EXPAND)
        sizer.Add(gsizer, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL)
        sizer2.Add(label2, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL)
        sizer2.Add(self.txt2, 1, wx.ALL | wx.EXPAND)
        sizer2.Add(self.btn_replace, 0, wx.ALL | wx.EXPAND)
        sizer2.Add(self.btn_replaceall, 0, wx.ALL | wx.EXPAND)

        self.Layout()
        self.Fit()

    def get_find_text(self):
        return str(self.txt.GetValue())

    def get_replace_text(self):
        return str(self.txt2.GetValue())


class PanelWithFindPanel(wx.Panel):
    def __init__(self, *args, **kwargs):
        wx.Panel.__init__(self, *args, **kwargs)

        self.find_panel = FindPanel(self, wx.ID_ANY)
        self._find_shown = False
        self.SetSizer(wx.BoxSizer(wx.VERTICAL))
        self._find_mode = 'forward'

    def ToggleFindPanel(self):
        if self._find_shown:
            nb = self.GetChildren()[1]
            stc = nb.GetCurrentPage()
            stc.SetFocus()
            self.GetSizer().Detach(self.find_panel)
            self._find_shown = False
        else:
            self.GetSizer().Add(self.find_panel, 0, wx.EXPAND)
            self._find_shown = True
        self.Layout()

    def get_findpanel_shown(self):
        return self._find_shown

    def onHitCL(self, evt):
        self.ToggleFindPanel()

    def find_forward(self):
        nb = self.GetChildren()[1]
        stc = nb.GetCurrentPage()
        txt = self.find_panel.get_find_text()

        l1, l2 = stc.GetSelection()
        for i in range(l2-l1):
            stc.CharRight()
        stc.SearchAnchor()
        flag = stc.SearchNext(0, txt)
        if flag != -1:
            l1, l2 = stc.GetSelection()
            stc.SetCurrentPos(l2)
            stc.SetSelection(l1, l2)
        return flag

    def find_backward(self):
        nb = self.GetChildren()[1]
        stc = nb.GetCurrentPage()
        txt = self.find_panel.get_find_text()

        stc.SearchAnchor()
        flag = stc.SearchPrev(0, txt)
        if flag != -1:
            l1, l2 = stc.GetSelection()
            stc.SetCurrentPos(l1)
            stc.SetSelection(l1, l2)
        return flag

    def onHitFW(self, evt):
        nb = self.GetChildren()[1]
        stc = nb.GetCurrentPage()
        flag = self.find_forward()
        stc.EnsureCaretVisible()
        self._find_mode = 'forward'
        evt.Skip()

    def onHitBW(self, evt):
        nb = self.GetChildren()[1]
        stc = nb.GetCurrentPage()
        flag = self.find_backward()
        stc.EnsureCaretVisible()
        self._find_mode = 'backward'
        evt.Skip()

    def onRunFind(self, evt):
        self.onHitFW(evt)
        evt.Skip()

    def replace_once(self):
        nb = self.GetChildren()[1]
        stc = nb.GetCurrentPage()
        txt = self.find_panel.get_replace_text()
        if len(txt) != 0:
            l1, l2 = stc.GetSelection()
            if l1 == l2:
                return False
            stc.Replace(l1, l2, txt)
            if self._find_mode == 'forward':
                flag = self.find_forward()
            else:
                flag = self.find_backward()
            stc.EnsureCaretVisible()
            return flag != -1
        return False

    def onReplace(self, evt):
        flag2 = self.replace_once()
        evt.Skip()

    def onReplaceAll(self, evt):
        nb = self.GetChildren()[1]
        stc = nb.GetCurrentPage()

        pos = stc.GetCurrentPos()

        txt = self.find_panel.get_find_text()
        rtxt = self.find_panel.get_replace_text()
        while 1:
            stc.SetCurrentPos(0)
            stc.SearchAnchor()
            flag = stc.SearchNext(0, txt)
            if flag == -1:
                break
            stc.ReplaceSelection(rtxt)

        stc.EnsureCaretVisible()
        evt.Skip()
